
"""Translator turning Scenic programs into Scenario objects.

The top-level interface to Scenic is provided by two functions:

* `scenarioFromString` -- compile a string of Scenic code;
* `scenarioFromFile` -- compile a Scenic file.

These output a `Scenario` object, from which scenes can be generated.
See the documentation for `Scenario` for details.

When imported, this module hooks the Python import system so that Scenic
modules can be imported using the ``import`` statement. This is primarily for the
translator's own use, but you could import Scenic modules from Python to
inspect them. Because Scenic uses Python's import system, the latter's rules
for finding modules apply, including the handling of packages.

Scenic is compiled in two main steps: translating the code into Python, and
executing the resulting Python module to generate a Scenario object encoding
the objects, distributions, etc. in the scenario. For details, see the function
`compileStream` below.
"""

import sys
import os
import io
import builtins
import time
import inspect
import types
import typing
import importlib
import importlib.abc
import importlib.util
import itertools
import pathlib
from collections import namedtuple, defaultdict
from contextlib import contextmanager

import tokenize
from tokenize import NAME, NL, NEWLINE, ENDMARKER, OP, NUMBER, COLON, COMMENT, ENCODING
from tokenize import LPAR, RPAR, LSQB, RSQB, RBRACE, COMMA, DOUBLESLASH, DOUBLESLASHEQUAL
from tokenize import AT, LEFTSHIFT, RIGHTSHIFT, VBAR, AMPER, TILDE, CIRCUMFLEX, STAR
from tokenize import LEFTSHIFTEQUAL, RIGHTSHIFTEQUAL, VBAREQUAL, AMPEREQUAL, CIRCUMFLEXEQUAL
from tokenize import INDENT, DEDENT, STRING, SEMI

import ast
from ast import parse, dump, NodeVisitor, NodeTransformer, copy_location, fix_missing_locations
from ast import Load, Store, Del, Name, Call, Tuple, BinOp, MatMult, BitAnd, BitOr, BitXor, LShift
from ast import RShift, Starred, Lambda, AnnAssign, Set, Str, Subscript, Index, IfExp
from ast import NamedExpr, Yield, YieldFrom, FunctionDef, Attribute, Constant, Assign, Expr
from ast import Return, Raise, If, UnaryOp, Not, ClassDef, Nonlocal, Global, Compare, Is, Try
from ast import Break, Continue, AsyncFunctionDef, Pass, While

from scenic.core.distributions import Samplable, needsSampling, toDistribution
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.object_types import _Constructible
import scenic.core.errors as errors
from scenic.core.errors import (TokenParseError, PythonParseError, ASTParseError,
								InvalidScenarioError)
import scenic.core.dynamics as dynamics
import scenic.core.pruning as pruning
import scenic.syntax.veneer as veneer

### THE TOP LEVEL: compiling a Scenic program

def scenarioFromString(string, params={}, model=None, scenario=None,
					   filename='<string>', cacheImports=False):
	"""Compile a string of Scenic code into a `Scenario`.

	The optional **filename** is used for error messages."""
	stream = io.BytesIO(string.encode())
	return scenarioFromStream(stream, params=params, model=model, scenario=scenario,
							  filename=filename, cacheImports=cacheImports)

def scenarioFromFile(path, params={}, model=None, scenario=None, cacheImports=False):
	"""Compile a Scenic file into a `Scenario`.

	Args:
		path (str): path to a Scenic file
		params (dict): global parameters to override
		model (str): Scenic module to use as world model
		scenario (str): if there are multiple scenarios in the file, which one
		  to use
		cacheImports (bool): Whether to cache any imported Scenic modules.
		  The default behavior is to not do this, so that subsequent attempts
		  to import such modules will cause them to be recompiled. If it is
		  safe to cache Scenic modules across multiple compilations, set this
		  argument to True. Then importing a Scenic module will have the same
		  behavior as importing a Python module.

	Returns:
		A `Scenario` object representing the Scenic scenario.
	"""
	if not os.path.exists(path):
		raise FileNotFoundError(path)
	fullpath = os.path.realpath(path)
	head, extension = os.path.splitext(fullpath)
	if not extension or extension[1:] not in scenicExtensions:
		ok = ', '.join(scenicExtensions)
		err = f'Scenic scenario does not have valid extension ({ok})'
		raise RuntimeError(err)
	directory, name = os.path.split(head)

	with open(path, 'rb') as stream:
		return scenarioFromStream(stream, params=params, model=model, scenario=scenario,
								  filename=fullpath, path=path, cacheImports=cacheImports)

def scenarioFromStream(stream, params={}, model=None, scenario=None,
					   filename='<stream>', path=None, cacheImports=False):
	"""Compile a stream of Scenic code into a `Scenario`."""
	# Compile the code as if it were a top-level module
	oldModules = list(sys.modules.keys())
	try:
		with topLevelNamespace(path) as namespace:
			compileStream(stream, namespace, params=params, model=model, filename=filename)
	finally:
		if not cacheImports:
			toRemove = []
			for name, module in sys.modules.items():
				if name not in oldModules and getattr(module, '_isScenicModule', False):
					toRemove.append(name)
			for name in toRemove:
				del sys.modules[name]
	# Construct a Scenario from the resulting namespace
	return constructScenarioFrom(namespace, scenario)

@contextmanager
def topLevelNamespace(path=None):
	"""Creates an environment like that of a Python script being run directly.

	Specifically, __name__ is '__main__', __file__ is the path used to invoke
	the script (not necessarily its absolute path), and the parent directory is
	added to the path so that 'import blobbo' will import blobbo from that
	directory if it exists there.
	"""
	directory = os.getcwd() if path is None else os.path.dirname(path)
	namespace = { '__name__': '__main__' }
	if path is not None:
		namespace['__file__'] = path
	sys.path.insert(0, directory)
	try:
		yield namespace
	finally:
		del sys.path[0]

def compileStream(stream, namespace, params={}, model=None, filename='<stream>'):
	"""Compile a stream of Scenic code and execute it in a namespace.

	The compilation procedure consists of the following main steps:

		1. Tokenize the input using the Python tokenizer.
		2. Partition the tokens into blocks separated by import statements.
		   This is done by the `partitionByImports` function.
		3. Translate Scenic constructions into valid Python syntax.
		   This is done by the `TokenTranslator`.
		4. Parse the resulting Python code into an AST using the Python parser.
		5. Modify the AST to achieve the desired semantics for Scenic.
		   This is done by the `translateParseTree` function.
		6. Compile and execute the modified AST.
		7. After executing all blocks, extract the global state (e.g. objects).
		   This is done by the `storeScenarioStateIn` function.
	"""
	if verbosity >= 2:
		veneer.verbosePrint(f'  Compiling Scenic module from {filename}...')
		startTime = time.time()
	# Tokenize input stream
	try:
		tokens = list(tokenize.tokenize(stream.readline))
	except tokenize.TokenError as e:
		line = e.args[1][0] if isinstance(e.args[1], tuple) else e.args[1]
		raise TokenParseError(line, filename, 'file ended during multiline string or expression')
	# Partition into blocks with all imports at the end (since imports could
	# pull in new constructor (Scenic class) definitions, which change the way
	# subsequent tokens are transformed)
	blocks = partitionByImports(tokens)
	veneer.activate(params, model, filename, namespace)
	newSourceBlocks = []
	try:
		# Execute preamble
		exec(compile(preamble, '<veneer>', 'exec'), namespace)
		namespace[namespaceReference] = namespace
		# Execute each block
		for blockNum, block in enumerate(blocks):
			# Find all custom constructors defined so far (possibly imported)
			constructors = findConstructorsIn(namespace)
			# Translate tokens to valid Python syntax
			startLine = max(1, block[0][2][0])
			translator = TokenTranslator(constructors, filename)
			newSource, allConstructors = translator.translate(block)
			trimmed = newSource[2*(startLine-1):]	# fix up blank lines used to align errors
			newSource = '\n'*(startLine-1) + trimmed
			newSourceBlocks.append(trimmed)
			if dumpTranslatedPython:
				print(f'### Begin translated Python from block {blockNum} of {filename}')
				print(newSource)
				print('### End translated Python')
			# Parse the translated source
			tree = parseTranslatedSource(newSource, filename)
			# Modify the parse tree to produce the correct semantics
			newTree, requirements = translateParseTree(tree, allConstructors, filename)
			if dumpFinalAST:
				print(f'### Begin final AST from block {blockNum} of {filename}')
				print(ast.dump(newTree, include_attributes=True))
				print('### End final AST')
			if dumpASTPython:
				try:
					import astor
				except ModuleNotFoundError as e:
					raise RuntimeError('dumping the Python equivalent of the AST'
									   'requires the astor package')
				print(f'### Begin Python equivalent of final AST from block {blockNum} of {filename}')
				print(astor.to_source(newTree, add_line_information=True))
				print('### End Python equivalent of final AST')
			# Compile the modified tree
			code = compileTranslatedTree(newTree, filename)
			# Execute it
			executeCodeIn(code, namespace)
		# Extract scenario state from veneer and store it
		storeScenarioStateIn(namespace, requirements)
	finally:
		veneer.deactivate()
	if verbosity >= 2:
		totalTime = time.time() - startTime
		veneer.verbosePrint(f'  Compiled Scenic module in {totalTime:.4g} seconds.')
	allNewSource = ''.join(newSourceBlocks)
	return code, allNewSource

### TRANSLATION PHASE ZERO: definitions of language elements not already in Python

## Options

dumpTranslatedPython = False
dumpFinalAST = False
dumpASTPython = False
verbosity = 0
usePruning = True

## Preamble
# (included at the beginning of every module to be translated;
# imports the implementations of the public language features)
preamble = """\
from scenic.syntax.veneer import *
"""

## Get Python names of various elements
## (for checking consistency between the translator and the veneer)

api = set(veneer.__all__)

## Functions used internally

createDefault = 'PropertyDefault'
scenarioClass = 'DynamicScenario'
behaviorClass = 'Behavior'
monitorClass = 'Monitor'
createTerminationAction = 'makeTerminationAction'
internalFunctions = {
	createDefault, scenarioClass,
	behaviorClass, createTerminationAction,
}

# sanity check: these functions actually exist
for imp in internalFunctions:
	assert imp in api, imp

## Built-in functions

builtinFunctions = { 'resample', 'verbosePrint', 'simulation', 'localPath' }

# sanity check: implementations of built-in functions actually exist
for imp in builtinFunctions:
	assert imp in api, imp

## Built-in names (values which cannot be overwritten)

globalParametersName = 'globalParameters'

builtinNames = { globalParametersName }

# sanity check: built-in names actually exist
for name in builtinNames:
	assert name in api, name

## Simple statements

paramStatement = 'param'
mutateStatement = 'mutate'

requireStatement = 'require'
softRequirement = 'require_soft'	# not actually a statement, but a marker for the parser
requireAlwaysStatement = ('require', 'always')
terminateWhenStatement = ('terminate', 'when')
terminateSimulationWhenStatement = ('terminate', 'simulation', 'when')
terminateAfterStatement = ('terminate', 'after')

actionStatement = 'take'			# statement invoking a primitive action
waitStatement = 'wait'				# statement invoking a no-op action
terminateStatement = 'terminate'	# statement ending the simulation
abortStatement = 'abort'			# statement ending a try-interrupt statement
invokeStatement = 'do'				# statement invoking a behavior or scenario

modelStatement = 'model'
namespaceReference = '_Scenic_module_namespace'		# used in the implementation of 'model'

simulatorStatement = 'simulator'

oneWordStatements = {	# TODO clean up
	paramStatement, mutateStatement, requireStatement,
	actionStatement, waitStatement, terminateStatement,
	abortStatement, invokeStatement, simulatorStatement
}
twoWordStatements = {
	requireAlwaysStatement, terminateWhenStatement, terminateAfterStatement,
}
threeWordStatements = { terminateSimulationWhenStatement }

threeWordIncipits = { tokens[:2]: tokens[2] for tokens in threeWordStatements } # TODO improve
for incipit, last in threeWordIncipits.items():
	for tok2 in twoWordStatements:
		assert tok2 != incipit, (tok2, last)

# statements implemented by functions
functionStatements = {
	requireStatement, paramStatement, mutateStatement,
	modelStatement, simulatorStatement
}
twoWordFunctionStatements = {
	requireAlwaysStatement, terminateWhenStatement, terminateAfterStatement,
}
threeWordFunctionStatements = { terminateSimulationWhenStatement }
def functionForStatement(tokens):
	return '_'.join(tokens) if isinstance(tokens, tuple) else tokens
def nameForStatement(tokens):
	return ' '.join(tokens) if isinstance(tokens, tuple) else tokens

statementForImp = {
	functionForStatement(s): nameForStatement(s)
	for s in functionStatements | twoWordStatements
}

# sanity check: implementations actually exist
for imp in functionStatements:
	assert imp in api, imp
for tokens in twoWordFunctionStatements:
	assert len(tokens) == 2
	imp = functionForStatement(tokens)
	assert imp in api, imp
for tokens in threeWordFunctionStatements:
	assert len(tokens) == 3
	imp = functionForStatement(tokens)
	assert imp in api, imp

# statements allowed inside behaviors
behavioralStatements = {
	requireStatement, actionStatement, waitStatement,
	terminateStatement, abortStatement, invokeStatement,
}
behavioralImps = { functionForStatement(s) for s in behavioralStatements }

# statements allowed inside scenario composition blocks
compositionalStatements = {
	requireStatement, waitStatement, terminateStatement, abortStatement,
	invokeStatement,
}
compositionalImps = { functionForStatement(s) for s in compositionalStatements }

# statements encoding requirements which need special handling
requirementStatements = (
	requireStatement, softRequirement,
	functionForStatement(requireAlwaysStatement),
	functionForStatement(terminateWhenStatement),
)

statementRaiseMarker = '_Scenic_statement_'

## Try-interrupt blocks

interruptWhenStatement = ('interrupt', 'when')

interruptExceptMarker = '_Scenic_interrupt_'

## Constructors and specifiers

# statement defining a new constructor (Scenic class);
# we still recognize 'constructor' for backwards-compatibility
constructorStatements = ('class', 'constructor')

Constructor = namedtuple('Constructor', ('name', 'bases'))

builtinSpecifiers = {
	# position
	('visible', 'from'): 'VisibleFrom',
	('offset', 'by'): 'OffsetBy',
	('offset', 'along'): 'OffsetAlongSpec',
	('at',): 'At',
	('in',): 'In',
	('on',): 'In',
	('beyond',): 'Beyond',
	('visible',): 'VisibleSpec',
	('left', 'of'): 'LeftSpec',
	('right', 'of'): 'RightSpec',
	('ahead', 'of'): 'Ahead',
	('behind',): 'Behind',
	('following',): 'Following',

	# heading
	('apparently', 'facing'): 'ApparentlyFacing',
	('facing', 'toward'): 'FacingToward',
	('facing',): 'Facing'
}

# sanity check: implementations of specifiers actually exist
for imp in builtinSpecifiers.values():
	assert imp in api, imp

builtinConstructors = {
	'Point': Constructor('Point', None),
	'OrientedPoint': Constructor('OrientedPoint', 'Point'),
	'Object': Constructor('Object', 'OrientedPoint')
}

# sanity check: built-in constructors actually exist
for const in builtinConstructors:
	assert const in api, const

## Other compound statements

behaviorStatement = 'behavior'		# statement defining a new behavior
monitorStatement = 'monitor'		# statement defining a new monitor

scenarioMarker = '_Scenic_scenario_'	# not a statement, but a marker for the parser

# Scenario blocks
setupBlock = 'setup'
composeBlock = 'compose'
scenarioBlocks = { setupBlock, composeBlock, 'precondition', 'invariant' }

## Prefix operators

prefixOperators = {
	('relative', 'position'): 'RelativePosition',
	('relative', 'heading'): 'RelativeHeading',
	('apparent', 'heading'): 'ApparentHeading',
	('distance', 'from'): 'DistanceFrom',
	('distance', 'to'): 'DistanceFrom',
	('angle', 'from'): 'AngleFrom',
	('angle', 'to'): 'AngleTo',
	('ego', '='): 'ego',
	('front', 'left'): 'FrontLeft',
	('front', 'right'): 'FrontRight',
	('back', 'left'): 'BackLeft',
	('back', 'right'): 'BackRight',
	('front', 'of'): 'Front',
	('back', 'of'): 'Back',
	('left', 'of'): 'Left',
	('right', 'of'): 'Right',
	('follow',): 'Follow',
	('visible',): 'Visible',
	('not', 'visible'): 'NotVisible',
}
assert all(1 <= len(op) <= 2 for op in prefixOperators)
prefixIncipits = { op[0] for op in prefixOperators }
assert not any(op in oneWordStatements for op in prefixIncipits)
assert not any(op in twoWordStatements for op in prefixOperators)

# sanity check: implementations of prefix operators actually exist
for imp in prefixOperators.values():
	assert imp in api, imp

## Modifiers and terminators

class ModifierInfo(typing.NamedTuple):
	name: str
	terminators: typing.Tuple[str]
	contexts: typing.Optional[typing.Tuple[str]] = ()

modifiers = (
	ModifierInfo('for', ('seconds', 'steps'), (invokeStatement,)),
	ModifierInfo('until', (), (invokeStatement,)),
)

modifierNames = {}
for mod in modifiers:
	assert mod.name not in modifierNames, mod
	modifierNames[mod.name] = mod

terminatorsForStatements = {
	functionForStatement(terminateAfterStatement): ('seconds', 'steps'),
}

## Infix operators

# pseudo-operator for encoding argument packages for (3+)-ary ops
packageToken = (RIGHTSHIFT, '>>')
packageNode = RShift

class InfixOp(typing.NamedTuple):
	syntax: str
	implementation: typing.Optional[str]
	arity: int
	token: typing.Tuple[int, str]
	node: ast.AST
	contexts: typing.Optional[typing.Tuple[str]] = ()

infixOperators = (
	# existing Python operators with new semantics
	InfixOp('@', 'Vector', 2, None, MatMult),

	# operators not in Python (in decreasing precedence order)
	InfixOp('at', 'FieldAt', 2, (LEFTSHIFT, '<<'), LShift),
	InfixOp('relative to', 'RelativeTo', 2, (AMPER, '&'), BitAnd),
	InfixOp('offset by', 'RelativeTo', 2, (AMPER, '&'), BitAnd),
	InfixOp('offset along', 'OffsetAlong', 3, (CIRCUMFLEX, '^'), BitXor),
	InfixOp('can see', 'CanSee', 2, (VBAR, '|'), BitOr),

	# just syntactic conveniences, not really operators
	InfixOp('from', None, 2, (COMMA, ','), None),
	InfixOp('for', None, 2, (COMMA, ','), None, ('Follow', 'Following')),
	InfixOp('to', None, 2, (COMMA, ','), None),
	InfixOp('by', None, 2, packageToken, None)
)

infixTokens = {}
infixImplementations = {}
infixIncipits = set()
for op in infixOperators:
	# if necessary, set up map from Scenic to Python syntax
	if op.token is not None:
		tokens = tuple(op.syntax.split(' '))
		assert 1 <= len(tokens) <= 2, op
		assert tokens not in infixTokens, op
		assert tokens not in twoWordStatements, op
		infixTokens[tokens] = op
		incipit = tokens[0]
		assert incipit not in oneWordStatements, op
		infixIncipits.add(incipit)
	# if necessary, set up map from Python to Scenic semantics
	imp = op.implementation
	if imp is not None:
		assert imp in api, op
		node = op.node
		if node in infixImplementations:	# two operators may have the same implementation
			oldArity, oldName = infixImplementations[node]
			assert op.arity == oldArity, (op, oldName)
			assert imp == oldName, (op, oldName)
		else:
			infixImplementations[node] = (op.arity, imp)
generalInfixOps = { tokens: op.token for tokens, op in infixTokens.items() if not op.contexts }

## Direct syntax replacements

replacements = {	# TODO police the usage of these? could yield bizarre error messages
	'of': tuple(),
	'deg': ((STAR, '*'), (NUMBER, '0.01745329252')),
	'ego': ((NAME, 'ego'), (LPAR, '('), (RPAR, ')')),
	'globalParameters': ((NAME, 'globalParameters'), (LPAR, '('), (RPAR, ')')),
}

twoWordReplacements = {
	('initial', 'scenario'): ((NAME, 'in_initial_scenario'), (LPAR, '('), (RPAR, ')')),
}

## Illegal and reserved syntax

illegalTokens = {
	LEFTSHIFT, RIGHTSHIFT, VBAR, AMPER, TILDE, CIRCUMFLEX,
	LEFTSHIFTEQUAL, RIGHTSHIFTEQUAL, VBAREQUAL, AMPEREQUAL, CIRCUMFLEXEQUAL,
	DOUBLESLASH, DOUBLESLASHEQUAL
}

# sanity check: stand-in tokens for infix operators must be illegal
for op in infixTokens.values():
	ttype = op.token[0]
	assert (ttype is COMMA or ttype in illegalTokens), op

illegalConstructs = {
	('async', 'def'),		# used to parse behaviors, so disallowed otherwise
}

keywords = (
	set(constructorStatements)
	| (oneWordStatements - {simulatorStatement})
	| internalFunctions
	| replacements.keys()
)

### TRANSLATION PHASE ONE: handling imports

## Meta path finder and loader for Scenic files

scenicExtensions = ('sc', 'scenic')

class ScenicMetaFinder(importlib.abc.MetaPathFinder):
	def find_spec(self, name, paths, target=None):
		if paths is None:
			paths = sys.path
			modname = name
		else:
			modname = name.rpartition('.')[2]
		for path in paths:
			for extension in scenicExtensions:
				filename = modname + '.' + extension
				filepath = os.path.join(path, filename)
				if os.path.exists(filepath):
					filepath = os.path.abspath(filepath)
					spec = importlib.util.spec_from_file_location(name, filepath,
						loader=ScenicLoader(filepath, filename))
					return spec
		return None

class ScenicLoader(importlib.abc.InspectLoader):
	def __init__(self, filepath, filename):
		self.filepath = filepath
		self.filename = filename

	def create_module(self, spec):
		return None

	def exec_module(self, module):
		# Read source file and compile it
		with open(self.filepath, 'r') as stream:
			source = stream.read()
		with open(self.filepath, 'rb') as stream:
			code, pythonSource = compileStream(stream, module.__dict__, filename=self.filepath)
		# Mark as a Scenic module
		module._isScenicModule = True
		# Save code, source, and translated source for later inspection
		module._code = code
		module._source = source
		module._pythonSource = pythonSource

		# If we're in the process of compiling another Scenic module, inherit
		# objects, parameters, etc. from this one
		if veneer.isActive():
			veneer.currentScenario._inherit(module._scenario)

	def is_package(self, fullname):
		return False

	def get_code(self, fullname):
		module = importlib.import_module(fullname)
		assert module._isScenicModule, module
		return module._code

	def get_source(self, fullname):
		module = importlib.import_module(fullname)
		assert module._isScenicModule, module
		return module._pythonSource

# register the meta path finder
sys.meta_path.insert(0, ScenicMetaFinder())

## Miscellaneous utilities

def partitionByImports(tokens):
	"""Partition the tokens into blocks ending with import statements.

	We avoid splitting top-level try-except statements, to allow the pattern of trying
	to import an optional module and catching an ImportError. If someone tries to define
	objects inside such a statement, woe unto them.
	"""
	blocks = []
	currentBlock = []
	duringImport = False
	seenTry = False
	haveImported = False
	finishLine = False
	parenLevel = 0
	tokens = Peekable(tokens)
	for token in tokens:
		startNewBlock = False
		if token.exact_type == LPAR:
			parenLevel += 1
		elif token.exact_type == RPAR:
			parenLevel -= 1
		if finishLine:
			if token.type in (NEWLINE, NL) and parenLevel == 0:
				finishLine = False
				if duringImport:
					duringImport = False
					haveImported = True
		else:
			assert not duringImport
			finishLine = True
			if token.type in (DEDENT, NEWLINE, NL, COMMENT, ENCODING):
				finishLine = False
				if (seenTry and token.type == DEDENT and token.start[1] == 0
					and peek(tokens).string not in ('except', 'else', 'finally')):
					seenTry = False
					haveImported = True 	# just in case the try contained imports
			elif token.start[1] == 0:
				if token.string in ('import', 'from', modelStatement):
					duringImport = True
				else:
					if haveImported:
						# could use new constructors; needs to be in a new block
						startNewBlock = True
					if token.string == 'try':
						seenTry = True
		if startNewBlock:
			blocks.append(currentBlock)
			currentBlock = [token]
			haveImported = False
		else:
			currentBlock.append(token)
	blocks.append(currentBlock)		# add last block
	return blocks

def findConstructorsIn(namespace):
	"""Find all constructors (Scenic classes) defined in a namespace."""
	constructors = []
	for name, value in namespace.items():
		if inspect.isclass(value) and issubclass(value, _Constructible):
			if name in builtinConstructors:
				continue
			parents = []
			for base in value.__bases__:
				if issubclass(base, _Constructible):
					parents.append(base.__name__)
			constructors.append(Constructor(name, parents))
	return constructors

### TRANSLATION PHASE TWO: translation at the level of tokens

class Peekable:
	"""Utility class to allow iterator lookahead."""
	def __init__(self, gen):
		self.gen = iter(gen)
		self.current = next(self.gen, None)

	def __iter__(self):
		return self

	def __next__(self):
		cur = self.current
		if cur is None:
			raise StopIteration
		self.current = next(self.gen, None)
		return cur

	def peek(self):
		return self.current

def peek(thing):
	return thing.peek()

class TokenTranslator:
	"""Translates a Scenic token stream into valid Python syntax.

	This is a stateful process because constructor (Scenic class) definitions
	change the way subsequent code is parsed.
	"""
	def __init__(self, constructors=(), filename='<unknown>'):
		self.constructors = dict(builtinConstructors)
		for constructor in constructors:
			name = constructor.name
			assert name not in self.constructors
			self.constructors[name] = constructor
		self.filename = filename

	def parseError(self, tokenOrLine, message):
		raise TokenParseError(tokenOrLine, self.filename, message)

	def createConstructor(self, name, parents):
		parents = tuple(parents)
		assert parents
		self.constructors[name] = Constructor(name, parents)

	def specifiersForConstructor(self, const):
		# Currently all specifiers can be used with any constructor;
		# I'm leaving this here in case we later allow custom specifiers to be inherited
		return builtinSpecifiers
		#name, parents = self.constructors[const]

	def translate(self, tokens):
		"""Do the actual translation of the token stream."""
		tokens = Peekable(tokens)
		newTokens = []
		functionStack = []
		context, startLevel = None, None
		specifiersIndented = False
		parenLevel = 0
		row, col = 0, 0		# position of next token to write out
		orow, ocol = 0, 0	# end of last token in the original source
		startOfLine = True
		startOfStatement = True 	# position where a (simple) statement could begin
		constructors = self.constructors

		for token in tokens:
			ttype = token.exact_type
			tstring = token.string
			skip = False
			endToken = token 	# token to advance past in column count
			movedUpTo = False

			def injectToken(tok, spaceAfter=0):
				"""Add a token to the output stream, trying to preserve spacing."""
				nonlocal row, col, movedUpTo
				if not movedUpTo:
					moveUpTo(token)
					moveBeyond(token)
					movedUpTo = True
				ty, string = tok[:2]
				if len(tok) >= 3:
					moveBeyond(tok)
					srow, scol = tok[2]
					erow, ecol = tok[3]
					width = ecol - scol
					height = erow - srow
				else:
					width = len(string)
					height = 0
				ncol = ecol if height > 0 else col + width
				newToken = (ty, string, (row, col), (row+height, ncol), '')
				newTokens.append(newToken)
				if ty in (NEWLINE, NL):
					row += 1
					col = 0
				elif height > 0:
					row += height
					col = ncol
				else:
					col += width + spaceAfter
			def moveUpTo(tok):
				nonlocal row, col, orow, ocol
				nrow, ncol = tok[2]
				if nrow > orow:
					row = nrow
					col = ncol
				else:
					gap = ncol - ocol
					assert gap >= 0, (tok, row, col, ocol)
					col += gap
			def moveBeyond(tok):
				nonlocal orow, ocol
				nrow, ncol = tok[3]
				if nrow > orow or (nrow == orow and ncol > ocol):
					orow = nrow
					ocol = ncol
			def advance(skip=True):
				nextToken = next(tokens)
				if skip:
					moveBeyond(nextToken)
				else:
					injectToken(nextToken)
				return nextToken
			def callFunction(function, argument=None, implementation=None):
				nonlocal skip, matched, functionStack
				functionStack.append((function, parenLevel))
				implementation = function if implementation is None else implementation
				injectToken((NAME, implementation))
				injectToken((LPAR, '('))
				if argument is not None:
					injectToken(argument)
					injectToken((COMMA, ','))
				skip = True
				matched = True
			def popFunction():
				nonlocal context, startLevel
				functionStack.pop()
				injectToken((RPAR, ')'))
				context, startLevel = (None, 0) if len(functionStack) == 0 else functionStack[-1]
			def wrapStatementCall():
				injectToken((NAME, 'raise'), spaceAfter=1)
				injectToken((NAME, statementRaiseMarker), spaceAfter=1)
				injectToken((NAME, 'from'), spaceAfter=1)

			# Catch Python operators that can't be used in Scenic
			if ttype in illegalTokens:
				self.parseError(token, f'illegal operator "{tstring}"')

			# Determine which operators are allowed in current context
			allowedPrefixOps = prefixOperators
			allowedInfixOps = dict()
			allowedModifiers = dict()
			allowedTerminators = set()
			inConstructorContext = False
			context, startLevel = functionStack[-1] if functionStack else (None, None)
			if parenLevel == startLevel:
				if context in constructors:
					inConstructorContext = True
					allowedPrefixOps = self.specifiersForConstructor(context)
				else:
					for opTokens, op in infixTokens.items():
						if not op.contexts or context in op.contexts:
							allowedInfixOps[opTokens] = op.token
					for name, mod in modifierNames.items():
						if not mod.contexts or context in mod.contexts:
							allowedModifiers[name] = mod.name
					if context in modifierNames:
						allowedTerminators = modifierNames[context].terminators
					elif context in terminatorsForStatements:
						allowedTerminators = terminatorsForStatements[context]
			else:
				allowedInfixOps = generalInfixOps

			# Parse next token
			if ttype == LPAR or ttype == LSQB:		# keep track of nesting level
				parenLevel += 1
			elif ttype == RPAR or ttype == RSQB:	# ditto
				parenLevel -= 1
			elif ttype == STRING:
				# special case for global parameters with quoted names:
				# transform "name"=value into "name", value
				if (len(functionStack) > 0 and functionStack[-1][0] == paramStatement
					and peek(tokens).string == '='):
					next(tokens)	# consume '='
					injectToken(token)
					injectToken((COMMA, ','))
					skip = True
			elif ttype == NAME:		# the interesting case: almost all new syntax falls in here
				# try to match 2-word language constructs
				matched = False
				nextToken = peek(tokens)		# lookahead so we can give 2-word ops precedence
				if nextToken is not None:
					endToken = nextToken	# tentatively; will be overridden if no match
					nextString = nextToken.string
					twoWords = (tstring, nextString)
					if startOfLine and tstring == 'for':	# TODO improve hack?
						matched = True
						endToken = token
					elif startOfLine and tstring in constructorStatements:	# class definition
						if nextToken.type != NAME or nextString in keywords:
							self.parseError(nextToken, f'invalid class name "{nextString}"')
						nextToken = next(tokens)	# consume name
						bases, scenicParents = [], []
						if peek(tokens).exact_type == LPAR:		# superclass specification
							next(tokens)
							nextToken = next(tokens)
							while nextToken.exact_type != RPAR:
								base = nextToken.string
								if nextToken.exact_type != NAME:
									self.parseError(nextToken, f'invalid superclass "{base}"')
								bases.append(base)
								if base in self.constructors:
									scenicParents.append(base)
								if peek(tokens).exact_type == COMMA:
									next(tokens)
								nextToken = next(tokens)
							if not scenicParents and tstring != 'class':
								self.parseError(nextToken,
									f'Scenic class definition with no Scenic superclasses')
						if peek(tokens).exact_type != COLON:
							self.parseError(peek(tokens), 'malformed class definition')
						if not bases:
							bases = scenicParents = ('Object',)		# default superclass
						if scenicParents:
							self.createConstructor(nextString, scenicParents)
						injectToken((NAME, 'class'), spaceAfter=1)
						injectToken((NAME, nextString))
						injectToken((LPAR, '('))
						injectToken((NAME, bases[0]))
						for base in bases[1:]:
							injectToken((COMMA, ','), spaceAfter=1)
							injectToken((NAME, base))
						injectToken((RPAR, ')'))
						skip = True
						matched = True
						endToken = nextToken
					elif startOfLine and tstring == 'scenario':		# scenario definition
						if nextToken.type != NAME:
							self.parseError(nextToken, f'invalid scenario name "{nextString}"')
						className = scenarioMarker + nextString
						injectToken((NAME, 'async'), spaceAfter=1)
						injectToken((NAME, 'def'), spaceAfter=1)
						injectToken((NAME, className))
						advance()	# consume name
						skip = True
						matched = True
					elif startOfLine and tstring == behaviorStatement:		# behavior definition
						if nextToken.type != NAME:
							self.parseError(nextToken, f'invalid behavior name "{nextString}"')
						injectToken((NAME, 'async'), spaceAfter=1)
						injectToken((NAME, 'def'), spaceAfter=1)
						skip = True
						matched = True
						endToken = token
					elif startOfLine and tstring == monitorStatement:		# monitor definition
						if nextToken.type != NAME:
							self.parseError(nextToken, f'invalid monitor name "{nextString}"')
						injectToken((NAME, 'async'), spaceAfter=1)
						injectToken((NAME, 'def'), spaceAfter=1)
						injectToken((NAME, dynamics.functionForMonitor(nextString)))
						injectToken((LPAR, '('))
						injectToken((RPAR, ')'))
						advance()	# consume name
						if peek(tokens).exact_type != COLON:
							self.parseError(nextToken, 'malformed monitor definition')
						skip = True
						matched = True
					elif twoWords in allowedPrefixOps:	# 2-word prefix operator
						callFunction(allowedPrefixOps[twoWords])
						advance()	# consume second word
					elif not startOfStatement and twoWords in allowedInfixOps:	# 2-word infix operator
						injectToken(allowedInfixOps[twoWords])
						advance()	# consume second word
						skip = True
						matched = True
					elif startOfLine and twoWords == interruptWhenStatement:
						# special case for interrupt when
						injectToken((NAME, 'except'), spaceAfter=1)
						callFunction(interruptExceptMarker)
						advance()	# consume second word
						matched = True
					elif startOfStatement and twoWords in threeWordIncipits:	# 3-word statement
						advance()	# consume second word
						endToken = advance()	# consume third word
						thirdWord = endToken.tstring
						expected = threeWordIncipits[twoWords]
						if thirdWord != expected:
							self.parseError(endToken,
							                f'expected "{expected}", got "{thirdWord}"')
						wrapStatementCall()
						function = functionForStatement(twoWords + (thirdWord,))
						callFunction(function)
						matched = True
					elif startOfStatement and twoWords in twoWordStatements:	# 2-word statement
						wrapStatementCall()
						function = functionForStatement(twoWords)
						callFunction(function)
						advance()   # consume second word
						matched = True
					elif inConstructorContext and tstring == 'with':	# special case for 'with' specifier
						callFunction('With', argument=(STRING, f'"{nextString}"'))
						advance()	# consume property name
					elif startOfStatement and tstring == requireStatement and nextString == '[':
						# special case for require[p]
						next(tokens)	# consume '['
						nextToken = next(tokens)
						if nextToken.exact_type != NUMBER:
							self.parseError(nextToken,
								'soft requirement must have constant probability')
						prob = nextToken.string
						if not 0 <= float(prob) <= 1:
							self.parseError(nextToken, 'probability must be between 0 and 1')
						nextToken = next(tokens)
						if nextToken.exact_type != RSQB:
							self.parseError(nextToken, 'malformed soft requirement')
						wrapStatementCall()
						callFunction(softRequirement, argument=(NUMBER, prob))
						endToken = nextToken
					elif twoWords in twoWordReplacements:	# 2-word direct replacement
						for tok in twoWordReplacements[twoWords]:
							injectToken(tok, spaceAfter=1)
						advance()	# consume second word
						skip = True
					elif twoWords in illegalConstructs:
						construct = ' '.join(twoWords)
						self.parseError(token,
							f'Python construct "{construct}" not allowed in Scenic')
				if not matched:
					# 2-word constructs don't match; try 1-word
					endToken = token
					oneWord = (tstring,)
					if oneWord in allowedPrefixOps:		# 1-word prefix operator
						callFunction(allowedPrefixOps[oneWord])
					elif not startOfStatement and oneWord in allowedInfixOps:	# 1-word infix operator
						injectToken(allowedInfixOps[oneWord])
						skip = True
					elif inConstructorContext:		# couldn't match any 1- or 2-word specifier
						self.parseError(token, f'unknown specifier "{tstring}"')
					elif not startOfStatement and tstring in allowedModifiers:
						injectToken((COMMA, ','))
						callFunction(tstring, argument=(STRING, f'"{tstring}"'),
									 implementation='Modifier')
						skip = True
					elif not startOfStatement and tstring in allowedTerminators:
						injectToken((COMMA, ','))
						injectToken((STRING, f'"{tstring}"'))
						popFunction()
						skip = True
					elif startOfStatement and tstring in oneWordStatements:		# 1-word statement
						wrapStatementCall()
						callFunction(tstring)
					elif token.start[1] == 0 and tstring == modelStatement:		# model statement
						components = []
						while peek(tokens).exact_type not in (COMMENT, NEWLINE):
							nextToken = next(tokens)
							if nextToken.exact_type != NAME and nextToken.string != '.':
								self.parseError(nextToken, 'invalid module name')
							components.append(nextToken.string)
						if not components:
							self.parseError(token, 'model statement is missing module name')
						components.append("'")
						literal = "'" + ''.join(components)
						callFunction(modelStatement, argument=(NAME, namespaceReference))
						injectToken((STRING, literal))
						skip = True
					elif startOfLine and tstring in scenarioBlocks:		# named block of scenario
						if peek(tokens).exact_type != COLON:
							self.parseError(peek(tokens), f'malformed "{tstring}" block')
						injectToken((NAME, 'async'), spaceAfter=1)
						injectToken((NAME, 'def'), spaceAfter=1)
						injectToken((NAME, tstring))
						injectToken((LPAR, '('))
						injectToken((RPAR, ')'))
						skip = True
					elif (tstring in self.constructors
						  and peek(tokens).exact_type not in (RPAR, RSQB, RBRACE, COMMA)):
						# instance definition
						callFunction(tstring)
					elif tstring in replacements:	# direct replacement
						for tok in replacements[tstring]:
							injectToken(tok, spaceAfter=1)
						skip = True
					elif startOfLine and tstring == 'from':		# special case to allow 'from X import Y'
						pass
					elif tstring in keywords:		# some malformed usage
						self.parseError(token, f'unexpected keyword "{tstring}"')
					elif tstring in illegalConstructs:
						self.parseError(token,
							f'Python construct "{tstring}" not allowed in Scenic')
					else:
						pass	# nothing matched; pass through unchanged to Python

			# Detect the end of function argument lists
			if len(functionStack) > 0:
				context, startLevel = functionStack[-1]
				while parenLevel < startLevel:		# we've closed all parens for the current function
					popFunction()
				inConstructor = any(context in constructors for context, sl in functionStack)
				if inConstructor and parenLevel == startLevel and ttype == COMMA:		# starting a new specifier
					while functionStack and context not in constructors:
						popFunction()
					# allow the next specifier to be on the next line, if indented
					injectToken(token)		# emit comma immediately
					skip = True
					nextToken = peek(tokens)
					specOnNewLine = False
					while nextToken.exact_type in (NEWLINE, NL, COMMENT, ENDMARKER):
						specOnNewLine = True
						if nextToken.exact_type == COMMENT:
							advance(skip=False)		# preserve comment
							nextToken = peek(tokens)
						if nextToken.exact_type not in (NEWLINE, NL):
							self.parseError(nextToken, 'comma with no specifier following')
						advance(skip=False)		# preserve newline
						nextToken = peek(tokens)
					if specOnNewLine and not specifiersIndented:
						nextToken = next(tokens)		# consume indent
						if nextToken.exact_type != INDENT:
							self.parseError(nextToken,
								'expected indented specifier (extra comma on previous line?)')
						injectToken(nextToken)
						specifiersIndented = True
				elif ttype in (NEWLINE, ENDMARKER, COMMENT, SEMI):	# end of line or statement
					if parenLevel != 0:
						self.parseError(token, 'unmatched parens/brackets')
					interrupt = False
					if functionStack and functionStack[0][0] == interruptExceptMarker:
						lastToken = newTokens[-1]
						if lastToken[1] != ':':
							self.parseError(nextToken, 'expected colon for interrupt')
						newTokens.pop()		# remove colon for now
						interrupt = True
					while len(functionStack) > 0:
						functionStack.pop()
						injectToken((RPAR, ')'))
					if interrupt:
						injectToken((COLON, ':'))

			# Output token unchanged, unless handled above
			if not skip:
				injectToken(token)
			else:
				moveBeyond(endToken)
			startOfLine = (ttype in (ENCODING, NEWLINE, NL, INDENT, DEDENT))
			startOfStatement = startOfLine or (ttype == SEMI)

		rewrittenSource = tokenize.untokenize(newTokens)
		if not isinstance(rewrittenSource, str):	# TODO improve?
			rewrittenSource = str(rewrittenSource, encoding='utf-8')
		return rewrittenSource, self.constructors

### TRANSLATION PHASE THREE: parsing of Python resulting from token translation

def parseTranslatedSource(source, filename):
	try:
		tree = parse(source, filename=filename)
		return tree
	except SyntaxError as e:
		raise PythonParseError(e) from None

### TRANSLATION PHASE FOUR: modifying the parse tree

temporaryName = '_Scenic_temporary_name'
behaviorArgName = '_Scenic_current_behavior'
checkInvariantsName = '_Scenic_check_invariants'
interruptPrefix = '_Scenic_interrupt'

abortFlag = Attribute(Name('BlockConclusion', Load()), 'ABORT', Load())
breakFlag = Attribute(Name('BlockConclusion', Load()), 'BREAK', Load())
continueFlag = Attribute(Name('BlockConclusion', Load()), 'CONTINUE', Load())
returnFlag = Attribute(Name('BlockConclusion', Load()), 'RETURN', Load())
finishedFlag = Attribute(Name('BlockConclusion', Load()), 'FINISHED', Load())

noArgs = ast.arguments(
	posonlyargs=[],
	args=[], vararg=None,
	kwonlyargs=[], kw_defaults=[],
	kwarg=None, defaults=[])
selfArg = ast.arguments(
	posonlyargs=[],
	args=[ast.arg(arg='self', annotation=None)], vararg=None,
	kwonlyargs=[], kw_defaults=[],
	kwarg=None, defaults=[])
tempArg = ast.arguments(
	posonlyargs=[],
	args=[ast.arg(arg=temporaryName, annotation=None)], vararg=None,
	kwonlyargs=[], kw_defaults=[],
	kwarg=None, defaults=[Constant(None)])
initialBehaviorArgs = [
	ast.arg(arg=behaviorArgName, annotation=None),
	ast.arg(arg='self', annotation=None)
]
onlyBehaviorArgs = ast.arguments(
	posonlyargs=[],
	args=initialBehaviorArgs, vararg=None,
	kwonlyargs=[], kw_defaults=[],
	kwarg=None, defaults=[])

class AttributeFinder(NodeVisitor):
	"""Utility class for finding all referenced attributes of a given name."""
	@staticmethod
	def find(target, node):
		af = AttributeFinder(target)
		af.visit(node)
		return af.attributes

	def __init__(self, target):
		super().__init__()
		self.target = target
		self.attributes = set()

	def visit_Attribute(self, node):
		val = node.value
		if isinstance(val, Name) and val.id == self.target:
			self.attributes.add(node.attr)
		self.visit(val)

class LocalFinder(NodeVisitor):
	"""Utility class for finding all local variables of a code block."""
	@staticmethod
	def findIn(block, ignoreTemporaries=True):
		lf = LocalFinder()
		for statement in block:
			lf.visit(statement)
		if ignoreTemporaries:
			names = set(name for name in lf.names if not name.startswith(temporaryName))
		else:
			names = lf.names
		return names - lf.globals - lf.nonlocals

	def __init__(self):
		self.names = set()
		self.globals = set()
		self.nonlocals = set()

	def visit_Global(self, node):
		self.globals.update(node.names)

	def visit_Nonlocal(self, node):
		self.nonlocals.update(node.names)

	def visit_FunctionDef(self, node):
		self.names.add(node.name)
		self.visit(node.args)
		for decorator in node.decorator_list:
			self.visit(decorator)
		if node.returns is not None:
			self.visit(node.returns)
		# do not visit body; it's another block

	def visit_Lambda(self, node):
		self.visit(node.args)
		# do not visit body; it's another block

	def visit_ClassDef(self, node):
		self.names.add(node.name)
		for child in itertools.chain(node.bases, node.keywords, node.decorator_list):
			self.visit(child)
		# do not visit body; it's another block

	def visit_Import(self, node):
		for alias in node.names:
			bound = alias.asname
			if bound is None:
				bound = alias.name
			self.names.add(bound)

	def visit_ImportFrom(self, node):
		self.visit_Import(node)

	def visit_Name(self, node):
		if isinstance(node.ctx, (Store, Del)):
			self.names.add(node.id)

	def visit_ExceptHandler(self, node):
		if node.name is not None:
			self.names.add(node.name)
		self.generic_visit(node)

class ASTSurgeon(NodeTransformer):
	def __init__(self, constructors, filename):
		super().__init__()
		self.constructors = set(constructors.keys())
		self.filename = filename
		self.requirements = []
		self.inRequire = False
		self.inCompose = False
		self.inBehavior = False
		self.inGuard = False
		self.inTryInterrupt = False
		self.inInterruptBlock = False
		self.inLoop = False
		self.usedBreakOrContinue = False
		self.callDepth = 0

	def parseError(self, node, message):
		raise ASTParseError(node, message, self.filename)

	def unpack(self, arg, expected, node):
		"""Unpack arguments to ternary (and up) infix operators."""
		assert expected > 0
		if isinstance(arg, BinOp) and isinstance(arg.op, packageNode):
			if expected == 1:
				self.parseError(node, 'gave too many arguments to infix operator')
			else:
				return self.unpack(arg.left, expected - 1, node) + [self.visit(arg.right)]
		elif expected > 1:
			self.parseError(node, 'gave too few arguments to infix operator')
		else:
			return [self.visit(arg)]

	def visit(self, node):
		if isinstance(node, ast.AST):
			return super().visit(node)
		elif isinstance(node, list):
			newStatements = []
			for statement in node:
				newStatement = self.visit(statement)
				if isinstance(newStatement, ast.AST):
					newStatements.append(newStatement)
				else:
					newStatements.extend(newStatement)
			return newStatements
		else:
			raise RuntimeError(f'unknown object {node} encountered during AST surgery')

	def visit_Name(self, node):
		if node.id in builtinNames:
			if not isinstance(node.ctx, Load):
				self.parseError(node, f'unexpected keyword "{node.id}"')
		return node

	def visit_BinOp(self, node):
		"""Convert infix operators to calls to the corresponding Scenic internal functions."""
		left = node.left
		right = node.right
		op = node.op
		if isinstance(op, packageNode):		# unexpected argument package
			self.parseError(node, 'unexpected keyword "by"')
		elif type(op) in infixImplementations:	# an operator with non-Python semantics
			arity, impName = infixImplementations[type(op)]
			implementation = Name(impName, Load())
			copy_location(implementation, node)
			assert arity >= 2
			args = [self.visit(left)] + self.unpack(right, arity-1, node)
			newNode = Call(implementation, args, [])
		else:	# all other operators have the Python semantics
			newNode = BinOp(self.visit(left), op, self.visit(right))
		return copy_location(newNode, node)

	def visit_Raise(self, node):
		"""Handle Scenic statements encoded as raise statements.

		In particular:
		  * wrap require statements with lambdas;
		  * handle primitive action invocations inside behaviors;
		  * call the veneer implementations of other statements.
		"""
		if not isinstance(node.exc, Name) or node.exc.id != statementRaiseMarker:
			return self.generic_visit(node)		# an ordinary raise statement
		assert isinstance(node.cause, Call)
		assert isinstance(node.cause.func, Name)
		node = node.cause 	# move to inner call
		func = node.func
		assert isinstance(func, Name)
		if self.inCompose and func.id not in compositionalImps:
			statement = statementForImp[func.id]
			self.parseError(node, f'"{statement}" cannot be used in a {composeBlock} block')
		if self.inBehavior and func.id not in behavioralImps:
			statement = statementForImp[func.id]
			self.parseError(node, f'"{statement}" cannot be used in a behavior')
		if func.id in requirementStatements:		# require, terminate when, etc.
			if func.id == softRequirement:
				func.id = requireStatement
				numArgs = 2
				if len(node.args) != 2:
					self.parseError(node, f'"require" takes exactly 1 argument')
			else:
				numArgs = 1
			self.validateSimpleCall(node, numArgs)
			cond = node.args[-1]
			assert not self.inRequire
			self.inRequire = True
			req = self.visit(cond)
			self.inRequire = False
			reqID = Constant(len(self.requirements), None)	# save ID number
			self.requirements.append(req)		# save condition for later inspection when pruning
			closure = Lambda(noArgs, req)		# enclose requirement in a lambda
			lineNum = Constant(node.lineno, None)			# save line number for error messages
			copy_location(closure, req)
			copy_location(lineNum, req)
			newArgs = [reqID, closure, lineNum]
			if numArgs == 2:		# get probability for soft requirements
				prob = node.args[0]
				assert isinstance(prob, Constant) and isinstance(prob.value, (float, int))
				newArgs.append(prob)
			return copy_location(Expr(Call(func, newArgs, [])), node)
		elif func.id == simulatorStatement:
			self.validateSimpleCall(node, 1)
			sim = self.visit(node.args[0])
			closure = copy_location(Lambda(noArgs, sim), sim)
			return copy_location(Expr(Call(func, [closure], [])), node)
		elif func.id == invokeStatement:		# Sub-behavior or sub-scenario statement
			seenModifier = False
			invoked = []
			args = []
			for arg in node.args:
				if (isinstance(arg, Call) and isinstance(arg.func, Name)
					and arg.func.id == 'Modifier'):
					if seenModifier:
						self.parseError(arg,
							f'incompatible qualifiers for "{invokeStatement}" statement')
					seenModifier = True
					assert len(arg.args) >= 2
					assert isinstance(arg.args[0], Constant)
					if arg.args[0].value == 'until':
						arg.args[1] = Lambda(noArgs, arg.args[1])
					args.append(self.visit(arg))
				elif seenModifier:
					self.parseError(arg, f'malformed "{invokeStatement}" statement')
				else:
					invoked.append(self.visit(arg))
			maxInvoked = 1 if self.inBehavior and not self.inCompose else None
			self.validateSimpleCall(node, (1, maxInvoked), onlyInBehaviors=True, args=invoked)
			subHandler = Attribute(Name(behaviorArgName, Load()), '_invokeSubBehavior', Load())
			subArgs = [Name('self', Load()), Tuple(invoked, Load())] + args
			subRunner = Call(subHandler, subArgs, [])
			return self.generateInvocation(node, subRunner, invoker=YieldFrom)
		elif func.id == actionStatement:		# Action statement
			if self.inCompose:
				self.parseError(func, f'cannot use "{actionStatement}" in a {composeBlock} block')
			self.validateSimpleCall(node, (1, None), onlyInBehaviors=True)
			action = Tuple(self.visit(node.args), Load())
			return self.generateInvocation(node, action)
		elif func.id == waitStatement:		# Wait statement
			self.validateSimpleCall(node, 0, onlyInBehaviors=True)
			return self.generateInvocation(node, Constant((), None))
		elif func.id == terminateStatement:		# Terminate statement
			self.validateSimpleCall(node, 0, onlyInBehaviors=True)
			termination = Call(Name(createTerminationAction, Load()),
							   [Constant(node.lineno, None)], [])
			return self.generateInvocation(node, termination)
		elif func.id == abortStatement:		# abort statement for try-interrupt statements
			if not self.inTryInterrupt:
				self.parseError(node, '"abort" outside of try-interrupt statement')
			self.validateSimpleCall(node, 0, onlyInBehaviors=True)
			return copy_location(Return(abortFlag), node)
		else:
			# statement implemented directly by a function; leave call intact
			newCall = self.visit_Call(node)
			return copy_location(Expr(newCall), node)

	def validateSimpleCall(self, node, numArgs, onlyInBehaviors=False, args=None):
		func = node.func
		name = func.id
		if onlyInBehaviors and not self.inBehavior:
			self.parseError(node, f'"{name}" can only be used in a behavior')
		args = node.args if args is None else args
		if isinstance(numArgs, tuple):
			assert len(numArgs) == 2
			low, high = numArgs
			if high is not None and len(args) > high:
				self.parseError(node, f'"{name}" takes at most {high} argument(s)')
			if len(args) < low:
				self.parseError(node, f'"{name}" takes at least {low} argument(s)')
		elif len(args) != numArgs:
			self.parseError(node, f'"{name}" takes exactly {numArgs} argument(s)')
		if len(node.keywords) != 0:
			self.parseError(node, f'"{name}" takes no keyword arguments')
		for arg in args:
			if isinstance(arg, Starred):
				self.parseError(node, f'argument unpacking cannot be used with "{name}"')

	def generateInvocation(self, node, actionlike, invoker=Yield):
		"""Generate an invocation of an action, behavior, or scenario."""
		invokeAction = Expr(invoker(actionlike))
		checkInvariants = Expr(Call(Name(checkInvariantsName, Load()), [], []))
		return [copy_location(invokeAction, node), copy_location(checkInvariants, node)]

	def visit_Try(self, node):
		"""Handle try-interrupt blocks."""
		interrupts = []
		exceptionHandlers = []
		for handler in node.handlers:
			ty = handler.type
			if (isinstance(ty, Call) and isinstance(ty.func, Name)
				and ty.func.id == interruptExceptMarker):
				assert handler.name is None
				if len(ty.args) != 1:
					self.parseError(handler, '"interrupt when" takes a single expression')
				interrupts.append((ty.args[0], handler.body))
			else:
				exceptionHandlers.append(handler)
		if not interrupts:		# an ordinary try-except block
			return self.generic_visit(node)

		# Add dead copy of all interrupts to ensure all local variables defined inside
		# the body and handlers are also defined as locals in the top-level function
		statements = []
		oldInTryInterrupt = self.inTryInterrupt
		if not self.inTryInterrupt:
			deadcopy = If(Constant(False), [node], [])
			statements.append(deadcopy)
			self.inTryInterrupt = True

		# Construct body
		oldInInterruptBlock, oldInLoop = self.inInterruptBlock, self.inLoop
		self.inInterruptBlock = True
		self.inLoop = False
		self.usedBreakOrContinue = False

		def makeInterruptBlock(name, body):
			newBody = self.visit(body)
			allLocals = LocalFinder.findIn(newBody)
			if allLocals:
				newBody.insert(0, Nonlocal(list(allLocals)))
			newBody.append(Return(finishedFlag))
			return FunctionDef(name, onlyBehaviorArgs, newBody, [], None)

		bodyName = f'{interruptPrefix}_body'
		statements.append(makeInterruptBlock(bodyName, node.body))

		# Construct interrupt handlers and condition checkers
		handlerNames, conditionNames = [], []
		for i, (condition, block) in enumerate(interrupts):
			handlerName = f'{interruptPrefix}_handler_{i}'
			handlerNames.append(handlerName)
			conditionName = f'{interruptPrefix}_condition_{i}'
			conditionNames.append(conditionName)

			statements.append(makeInterruptBlock(handlerName, block))
			self.inGuard = True
			checker = Lambda(noArgs, self.visit(condition))
			self.inGuard = False
			defChecker = Assign([Name(conditionName, Store())], checker)
			statements.append(defChecker)

		self.inInterruptBlock, self.inLoop = oldInInterruptBlock, oldInLoop
		self.inTryInterrupt = oldInTryInterrupt

		# Prepare tuples of interrupt conditions and handlers
		# (in order from high priority to low, so reversed relative to the syntax)
		conditions = Tuple([Name(n, Load()) for n in reversed(conditionNames)], Load())
		handlers = Tuple([Name(n, Load()) for n in reversed(handlerNames)], Load())

		# Construct code to execute the try-interrupt statement
		args = [
			Name(behaviorArgName, Load()),
			Name('self', Load()),
			Name(bodyName, Load()),
			conditions,
			handlers
		]
		callRuntime = Call(Name('runTryInterrupt', Load()), args, [])
		runTI = Assign([Name(temporaryName, Store())], YieldFrom(callRuntime))
		statements.append(runTI)
		result = Name(temporaryName, Load())
		if self.usedBreakOrContinue:
			test = Compare(result, [Is()], [breakFlag])
			statements.append(If(test, [Break()], []))
			test = Compare(result, [Is()], [continueFlag])
			statements.append(If(test, [Continue()], []))
		test = Compare(result, [Is()], [returnFlag])
		retCheck = If(test, [Return(Attribute(result, 'return_value', Load()))], [])
		statements.append(retCheck)

		# Construct overall try-except statement
		if exceptionHandlers or node.finalbody:
			newTry = Try(statements,
						 [self.visit(handler) for handler in exceptionHandlers],
						 self.visit(node.orelse),
						 self.visit(node.finalbody))
			return copy_location(newTry, node)
		else:
			return statements

	def visit_For(self, node):
		old = self.inLoop
		self.inLoop = True
		newNode = self.generic_visit(node)
		self.inLoop = old
		return newNode

	def visit_While(self, node):
		old = self.inLoop
		self.inLoop = True
		newNode = self.generic_visit(node)
		self.inLoop = old
		return newNode

	def visit_FunctionDef(self, node):
		oldInLoop, oldInInterruptBlock = self.inLoop, self.inInterruptBlock
		self.inLoop, self.inInterruptBlock = False, False
		newNode = self.generic_visit(node)
		self.inLoop, self.inInterruptBlock = oldInLoop, oldInInterruptBlock
		return newNode

	def visit_Break(self, node):
		if self.inInterruptBlock and not self.inLoop:
			self.usedBreakOrContinue = True
			newNode = Return(breakFlag)
			return copy_location(newNode, node)
		else:
			return self.generic_visit(node)

	def visit_Continue(self, node):
		if self.inInterruptBlock and not self.inLoop:
			self.usedBreakOrContinue = True
			newNode = Return(continueFlag)
			return copy_location(newNode, node)
		else:
			return self.generic_visit(node)

	def visit_Return(self, node):
		if self.inInterruptBlock:
			value = Constant(None) if node.value is None else node.value
			ret = Return(Call(returnFlag, [value], []))
			return copy_location(ret, node)
		else:
			return self.generic_visit(node)

	def visit_Call(self, node):
		"""Handle Scenic syntax and semantics of function calls.

		In particular:
		  * unpack argument packages for operators;
		  * check for iterable unpacking applied to distributions.
		"""
		func = node.func
		newArgs = []
		# Translate arguments, unpacking any argument packages
		self.callDepth += 1
		wrappedStar = False
		for arg in node.args:
			if isinstance(arg, BinOp) and isinstance(arg.op, packageNode):
				newArgs.extend(self.unpack(arg, 2, node))
			elif isinstance(arg, Starred) and not self.inBehavior:
				wrappedStar = True
				checkedVal = Call(Name('wrapStarredValue', Load()),
								  [self.visit(arg.value), Constant(arg.value.lineno, None)],
								  [])
				newArgs.append(Starred(checkedVal, Load()))
			else:
				newArgs.append(self.visit(arg))
		newKeywords = [self.visit(kwarg) for kwarg in node.keywords]
		newFunc = self.visit(func)
		self.callDepth -= 1
		if wrappedStar:
			newNode = Call(Name('callWithStarArgs', Load()), [newFunc] + newArgs, newKeywords)
		else:
			newNode = Call(newFunc, newArgs, newKeywords)
		newNode = copy_location(newNode, node)
		ast.fix_missing_locations(newNode)
		return newNode

	def visit_AsyncFunctionDef(self, node):
		"""Process Scenic constructs parsed as async function definitions.

		These include:
			* scenario definitions;
			* behavior (and monitor) definitions.
		"""
		if node.name.startswith(scenarioMarker):	# scenario definition
			return self.transformScenarioDefinition(node)
		else:
			return self.transformBehaviorDefinition(node)

	def transformScenarioDefinition(self, node):
		if self.inCompose:
			self.parseError(node, f'cannot define a scenario inside a {composeBlock} block')
		if self.inBehavior:
			self.parseError(node, 'cannot define a scenario inside a behavior')

		# Set up arguments for setup block (also inherited by compose block)
		args = node.args
		args.args = initialBehaviorArgs + args.args
		args = self.visit(args)

		# Extract named blocks from scenario body, if any
		simple = False	# simple scenario with no blocks
		setup, compose = None, None
		preconditions, invariants = [], []
		for statement in node.body:
			if isinstance(statement, AsyncFunctionDef):
				if statement.name == setupBlock:
					if setup:
						self.parseError(statement,
							f'scenario contains multiple "{setupBlock}" blocks')
					setup = statement
				elif statement.name == composeBlock:
					if compose:
						self.parseError(statement,
							f'scenario contains multiple "{composeBlock}" blocks')
					compose = statement
				elif statement.name in ('precondition', 'invariant'):
					if len(statement.body) != 1 or not isinstance(statement.body[0], Expr):
						self.parseError(statement.body[0], f'malformed precondition/invariant')
					self.inGuard = True
					test = self.visit(statement.body[0].value)
					self.inGuard = False
					assert isinstance(test, ast.AST)
					if statement.name == 'precondition':
						preconditions.append(test)
					else:
						invariants.append(test)
				else:
					simple = True
			else:
				# scenario contains actual code; assume it is simple
				simple = True
		if simple:	# simple scenario: entire body is implicitly the setup block
			if setup:
				self.parseError(setup, f'simple scenario cannot have a "{setupBlock}" block')
			if compose:
				self.parseError(compose, f'simple scenario cannot have a "{composeBlock}" block')
			if preconditions or invariants:
				self.parseError(node, f'simple scenario cannot have preconditions/invariants')

		# Construct compose block
		if compose or preconditions or invariants:
			self.inCompose = self.inBehavior = True
			preamble = self.makeBehaviorPreamble(preconditions, invariants)
			if compose:
				body = self.visit(compose.body)
				if setup:
					# make all locals in compose block actually use scope of setup block
					# (the compose definition will be placed inside the setup function)
					setupLocals = LocalFinder.findIn(setup.body)
					composeLocals = LocalFinder.findIn(body)
					commonLocals = setupLocals & composeLocals
					if commonLocals:
						body.insert(0, Nonlocal(list(commonLocals)))
			else:
				# generate no-op compose block to ensure invariants are checked
				wait = self.generateInvocation(node, Constant((), None))
				body = [While(Constant(True, None), wait, [])]
				compose = node 	# for copy_location below
			newBody = preamble + body
			newDef = FunctionDef('_compose', noArgs, newBody, [], None)
			self.inCompose = self.inBehavior = False
			compose = copy_location(newDef, compose)
		else:
			compose = Assign([Name('_compose', Store())], Constant(None, None))

		# Construct setup block
		if setup:
			oldBody = setup.body
			oldLoc = setup
		elif simple:
			oldBody = node.body
			oldLoc = node
		else:
			oldBody = []
			oldLoc = node
		getLocals = Call(Name('locals', Load()), [], [])
		returnStmt = Return(getLocals)
		newBody = self.visit(oldBody) + [compose, returnStmt]
		newDef = FunctionDef('_setup', args, newBody,
							 [Name(id='staticmethod', ctx=Load())], None)
		setup = copy_location(newDef, oldLoc)

		# Assemble scenario definition
		name = node.name[len(scenarioMarker):]
		body = [setup]
		newDef = ClassDef(name, [Name(scenarioClass, Load())], [], body, [])
		return copy_location(newDef, node)

	def transformBehaviorDefinition(self, node):
		if self.inCompose:
			self.parseError(node, f'cannot define a behavior inside a {composeBlock} block')
		if self.inBehavior:
			self.parseError(node, 'cannot define a behavior inside a behavior')

		# process body
		self.inBehavior = True
		oldInLoop = self.inLoop
		self.inLoop = False
		preconditions = []
		invariants = []
		newStatements = []
		# preserve docstring, if any
		if (isinstance(node.body[0], Expr) and isinstance(node.body[0].value, Constant)
		    and isinstance(node.body[0].value.value, str)):
			docstring = [node.body[0]]
			oldStatements = node.body[1:]
		else:
			docstring = []
			oldStatements = node.body
		# find precondition and invariant definitions
		for statement in oldStatements:
			if (isinstance(statement, AnnAssign)
				and isinstance(statement.target, Name)
				and statement.value is None
				and statement.simple):
				group = None
				if statement.target.id == 'precondition':
					group = preconditions
				elif statement.target.id == 'invariant':
					group = invariants
				else:
					self.parseError(statement, 'unknown type of behavior attribute')
				self.inGuard = True
				test = self.visit(statement.annotation)
				self.inGuard = False
				assert isinstance(test, ast.AST)
				group.append(test)
			else:
				newStatement = self.visit(statement)
				if isinstance(newStatement, ast.AST):
					newStatements.append(newStatement)
				else:
					newStatements.extend(newStatement)

		newBody = self.makeBehaviorPreamble(preconditions, invariants) + newStatements
		self.inBehavior = False
		self.inLoop = oldInLoop

		# add private current behavior argument and implicit 'self' argument
		newArgs = self.visit(node.args)
		newArgs.args = initialBehaviorArgs + newArgs.args

		# convert to class definition
		decorators = [self.visit(decorator) for decorator in node.decorator_list]
		genDefn = FunctionDef('makeGenerator', newArgs, newBody, decorators, node.returns)
		classBody = docstring + [genDefn]
		name = node.name
		if dynamics.isAMonitorName(name):
			superclass = monitorClass
			name = dynamics.monitorName(name)
		else:
			superclass = behaviorClass
		newDefn = ClassDef(name, [Name(superclass, Load())], [], classBody, [])

		return copy_location(newDefn, node)

	def makeBehaviorPreamble(self, preconditions, invariants):
		# generate precondition checks
		precondChecks = []
		for precondition in preconditions:
			call = Call(Name('PreconditionViolation', Load()),
			            [Name(behaviorArgName, Load()), Constant(precondition.lineno, None)],
			            [])
			throw = Raise(exc=call, cause=None)
			check = If(test=UnaryOp(Not(), precondition), body=[throw], orelse=[])
			precondChecks.append(copy_location(check, precondition))
		# generate invariant checker
		invChecks = []
		for invariant in invariants:
			call = Call(Name('InvariantViolation', Load()),
			            [Name(behaviorArgName, Load()), Constant(invariant.lineno, None)],
			            [])
			throw = Raise(exc=call, cause=None)
			check = If(test=UnaryOp(Not(), invariant), body=[throw], orelse=[])
			invChecks.append(copy_location(check, invariant))
		defineChecker = FunctionDef(checkInvariantsName, tempArg,
									invChecks + [Return(Name(temporaryName, Load()))], [], None)
		spot = Attribute(Name(behaviorArgName, Load()), 'checkInvariants', Store())
		saveChecker = Assign([spot], Name(checkInvariantsName, Load()))
		callChecker = Expr(Call(Name(checkInvariantsName, Load()), [], []))
		# assemble function body preamble
		preamble = precondChecks + [defineChecker, saveChecker, callChecker]
		return preamble

	def visit_Yield(self, node):
		if self.inCompose:
			self.parseError(node, f'"yield" is not allowed inside a {composeBlock} block')
		if self.inBehavior:
			self.parseError(node, '"yield" is not allowed inside a behavior')
		return self.generic_visit(node)

	def visit_YieldFrom(self, node):
		if self.inCompose:
			self.parseError(node, f'"yield from" is not allowed inside a {composeBlock} block')
		if self.inBehavior:
			self.parseError(node, '"yield from" is not allowed inside a behavior')
		return self.generic_visit(node)

	def visit_ClassDef(self, node):
		"""Handle Scenic constructs parsed as class definitions.

		In particular:
		  * transform Scenic class definitions into ordinary Python ones;
		  * leave ordinary Python class definitions alone.
		"""
		if node.name in self.constructors:		# Scenic class definition
			return self.transformScenicClass(node)
		else:		# ordinary Python class
			for base in node.bases:
				name = None
				if isinstance(base, Call):
					name = base.func.id
				elif isinstance(base, Name):
					name = base.id
				if name is not None and name in self.constructors:
					self.parseError(node,
									f'Python class {node.name} derives from Scenic class {name}')
			return self.generic_visit(node)

	def transformScenicClass(self, node):
		"""Process property defaults for Scenic classes."""
		newBody = []
		for child in node.body:
			child = self.visit(child)
			if isinstance(child, AnnAssign):	# default value for property
				origValue = child.annotation
				target = child.target
				# extract any attributes for this property
				metaAttrs = []
				if isinstance(target, Subscript):
					sl = target.slice
					if not isinstance(sl, Index):
						self.parseError(sl, 'malformed attributes for property default')
					sl = sl.value
					if isinstance(sl, Name):
						metaAttrs.append(sl.id)
					elif isinstance(sl, Tuple):
						for elt in sl.elts:
							if not isinstance(elt, Name):
								self.parseError(elt,
									'malformed attributes for property default')
							metaAttrs.append(elt.id)
					else:
						self.parseError(sl, 'malformed attributes for property default')
					newTarget = Name(target.value.id, Store())
					copy_location(newTarget, target)
					target = newTarget
				# find dependencies of the default value
				properties = AttributeFinder.find('self', origValue)
				# create default value object
				args = [
					Set([Str(prop) for prop in properties]),
					Set([Str(attr) for attr in metaAttrs]),
					Lambda(selfArg, origValue)
				]
				value = Call(Name(createDefault, Load()), args, [])
				copy_location(value, origValue)
				newChild = AnnAssign(
					target=target, annotation=value,
					value=None, simple=True)
				child = copy_location(newChild, child)
			newBody.append(child)
		node.body = newBody
		return node

def translateParseTree(tree, constructors, filename):
	"""Modify the Python AST to produce the desired Scenic semantics."""
	surgeon = ASTSurgeon(constructors, filename)
	tree = fix_missing_locations(surgeon.visit(tree))
	return tree, surgeon.requirements

### TRANSLATION PHASE FIVE: AST compilation

def compileTranslatedTree(tree, filename):
	try:
		return compile(tree, filename, 'exec')
	except SyntaxError as e:
		raise PythonParseError(e) from None

### TRANSLATION PHASE SIX: Python execution

def executeCodeIn(code, namespace):
	"""Execute the final translated Python code in the given namespace."""
	exec(code, namespace)

### TRANSLATION PHASE SEVEN: scenario construction

def storeScenarioStateIn(namespace, requirementSyntax):
	"""Post-process an executed Scenic module, extracting state from the veneer."""

	# Save requirement syntax and other module-level information
	moduleScenario = veneer.currentScenario
	factory = veneer.simulatorFactory
	bns = gatherBehaviorNamespacesFrom(moduleScenario._behaviors)
	def handle(scenario):
		scenario._requirementSyntax = requirementSyntax
		if isinstance(scenario, type):
			scenario._simulatorFactory = staticmethod(factory)
		else:
			scenario._simulatorFactory = factory
		scenario._behaviorNamespaces = bns
	handle(moduleScenario)
	namespace['_scenarios'] = tuple(veneer.scenarios)
	for scenarioClass in veneer.scenarios:
		handle(scenarioClass)

	# Extract requirements, scan for relations used for pruning, and create closures
	# (only for top-level scenario; modular scenarios will be handled when instantiated)
	moduleScenario._compileRequirements()

	# Save global parameters
	for name, value in veneer._globalParameters.items():
		if needsLazyEvaluation(value):
			raise InvalidScenarioError(f'parameter {name} uses value {value}'
									   ' undefined outside of object definition')
	for scenario in veneer.scenarios:
		scenario._bindGlobals(veneer._globalParameters)
	moduleScenario._bindGlobals(veneer._globalParameters)

	namespace['_scenario'] = moduleScenario

def gatherBehaviorNamespacesFrom(behaviors):
	"""Gather any global namespaces which could be referred to by behaviors.

	We'll need to rebind any sampled values in them at runtime.
	"""
	behaviorNamespaces = {}
	def registerNamespace(modName, ns):
		if modName not in behaviorNamespaces:
			behaviorNamespaces[modName] = ns
		else:
			assert behaviorNamespaces[modName] is ns
			return
		for name, value in ns.items():
			if (isinstance(value, types.ModuleType)
				and getattr(value, '_isScenicModule', False)):
				registerNamespace(value.__name__, value.__dict__)
			else:
				# Convert values requiring sampling to Distributions
				dval = toDistribution(value)
				if dval is not value:
					ns[name] = dval
	for behavior in behaviors:
		modName = behavior.__module__
		globalNamespace = behavior.makeGenerator.__globals__
		registerNamespace(modName, globalNamespace)
	return behaviorNamespaces

def constructScenarioFrom(namespace, scenarioName=None):
	"""Build a Scenario object from an executed Scenic module."""
	modularScenarios = namespace['_scenarios']
	if scenarioName:
		ty = namespace.get(scenarioName, None)
		if not (isinstance(ty, type) and issubclass(ty, dynamics.DynamicScenario)):
			raise RuntimeError(f'no scenario "{scenarioName}" found')
		if ty._requiresArguments():
			raise RuntimeError(f'cannot instantiate scenario "{scenarioName}"'
			                   ' with no arguments') from None

		dynScenario = ty()
	elif len(modularScenarios) > 1:
		raise RuntimeError('multiple choices for scenario to run '
		                   '(specify using the --scenario option)')
	elif modularScenarios and not modularScenarios[0]._requiresArguments():
		dynScenario = modularScenarios[0]()
	else:
		dynScenario = namespace['_scenario']

	if not dynScenario._prepared:	# true for all except top-level scenarios
		# Execute setup block (if any) to create objects and requirements;
		# extract any requirements and scan for relations used for pruning
		dynScenario._prepare()
	scenario = dynScenario._toScenario(namespace)

	# Prune infeasible parts of the space
	if usePruning:
		pruning.prune(scenario, verbosity=verbosity)

	return scenario



"""Support for dynamic behaviors and modular scenarios."""

from collections import defaultdict
import enum
import inspect
import itertools
import types

from scenic.core.distributions import Samplable, toDistribution, needsSampling
from scenic.core.errors import RuntimeParseError, InvalidScenarioError
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.requirements import (RequirementType, PendingRequirement,
                                      DynamicRequirement)
from scenic.core.simulators import (RejectSimulationException, EndSimulationAction,
                                    EndScenarioAction)
from scenic.core.utils import argsToString
from scenic.core.workspaces import Workspace

# Scenarios

class Invocable:
    def __init__(self, *args, **kwargs):
        if veneer.evaluatingGuard:
            raise RuntimeParseError(
                'tried to invoke behavior/scenario from inside guard or interrupt condition')
        self._runningIterator = None

    def _invokeSubBehavior(self, agent, subs, modifier=None):
        if modifier:
            if modifier.name == 'for':  # do X for Y [seconds | steps]
                timeLimit = modifier.value
                if not isinstance(timeLimit, (float, int)):
                    raise RuntimeParseError('"do X for Y" with Y not a number')
                assert modifier.terminator in (None, 'seconds', 'steps')
                if modifier.terminator != 'steps':
                    timeLimit /= veneer.currentSimulation.timestep
                startTime = veneer.currentSimulation.currentTime
                condition = lambda: veneer.currentSimulation.currentTime - startTime >= timeLimit
            elif modifier.name == 'until':  # do X until Y
                condition = modifier.value
            else:
                raise RuntimeError(f'internal parsing error: impossible modifier {modifier}')

            def body(behavior, agent):
                yield from self._invokeInner(agent, subs)
            handler = lambda behavior, agent: BlockConclusion.ABORT
            yield from runTryInterrupt(self, agent, body, [condition], [handler])
        else:
            yield from self._invokeInner(agent, subs)

    def _invokeInner(self, agent, subs):
        raise NotImplementedError

class DynamicScenario(Invocable):
    def __init_subclass__(cls, *args, **kwargs):
        veneer.registerDynamicScenarioClass(cls)

    _requirementSyntax = None   # overridden by subclasses
    _simulatorFactory = None
    _globalParameters = None

    def __init__(self, *args, **kwargs):
        super().__init__()
        self._args = args
        self._kwargs = kwargs

        self._ego = None
        self._objects = []      # ordered for reproducibility
        self._externalParameters = []
        self._pendingRequirements = defaultdict(list)
        self._requirements = []
        self._requirementDeps = set()   # things needing to be sampled to evaluate the requirements

        self._agents = []
        self._monitors = []
        self._behaviors = []
        self._alwaysRequirements = []
        self._terminationConditions = []
        self._terminateSimulationConditions = []

        self._subScenarios = []
        self._endWithBehaviors = False
        self._timeLimit = None
        self._timeLimitIsInSeconds = False
        self._prepared = False
        self._dummyNamespace = None

        self._timeLimitInSteps = None   # computed at simulation time
        self._elapsedTime = 0

    @classmethod
    def _dummy(cls, filename, namespace):
        scenario = cls()
        scenario._setup = None
        scenario._compose = None
        scenario._filename = filename   # for debugging
        scenario._prepared = True
        scenario._dummyNamespace = namespace
        return scenario

    @classmethod
    def _requiresArguments(cls):
        """Whether this scenario cannot be instantiated without arguments."""
        if cls._setup:
            func = cls._setup
        elif cls._compose:
            func = cls._compose
        else:
            return True
        sig = inspect.signature(func)
        try:
            sig.bind(None, None)    # first two arguments are added internally by Scenic
            return False
        except TypeError:
            return True

    @property
    def ego(self):
        return self._ego

    @property
    def objects(self):
        return tuple(self._objects)

    def _bindTo(self, scene):
        self._ego = scene.egoObject
        self._objects = list(scene.objects)
        self._agents = [obj for obj in scene.objects if obj.behavior is not None]
        self._alwaysRequirements = scene.alwaysRequirements
        self._terminationConditions = scene.terminationConditions
        self._terminateSimulationConditions = scene.terminateSimulationConditions

    def _prepare(self):
        assert not self._prepared
        self._prepared = True

        veneer.prepareScenario(self)
        with veneer.executeInScenario(self, inheritEgo=True):
            # Execute setup block
            if self._setup is not None:
                locs = self._setup(self, None, *self._args, **self._kwargs)
                self.__dict__.update(locs)  # save locals as if they were in class scope
                veneer.finishScenarioSetup(self)

        # Extract requirements, scan for relations used for pruning, and create closures
        self._compileRequirements()

    @classmethod
    def _bindGlobals(cls, globs):
        cls._globalParameters = globs

    def _start(self):
        assert self._prepared

        # Compute time limit now that we know the simulation timestep
        self._elapsedTime = 0
        self._timeLimitInSteps = self._timeLimit
        if self._timeLimitIsInSeconds:
            self._timeLimitInSteps /= veneer.currentSimulation.timestep

        veneer.startScenario(self)
        with veneer.executeInScenario(self):
            # Start compose block
            if self._compose is not None:
                if not inspect.isgeneratorfunction(self._compose):
                    from scenic.syntax.translator import composeBlock
                    raise RuntimeParseError(f'"{composeBlock}" does not invoke any scenarios')
                self._runningIterator = self._compose()

            # Initialize behavior coroutines of agents
            for agent in self._agents:
                assert isinstance(agent.behavior, Behavior), agent.behavior
                agent.behavior.start(agent)
            # Initialize monitor coroutines
            for monitor in self._monitors:
                monitor.start()

    def _step(self):
        # Check if we have reached the time limit, if any
        if self._timeLimitInSteps is not None and self._elapsedTime >= self._timeLimitInSteps:
            return self._stop('reached time limit')
        self._elapsedTime += 1

        # Execute compose block, if any
        composeDone = False
        if self._runningIterator is None:
            composeDone = True      # compose block ended in an earlier step
        else:
            with veneer.executeInScenario(self):
                try:
                    result = self._runningIterator.send(None)
                    if isinstance(result, (EndSimulationAction, EndScenarioAction)):
                        return self._stop(result)
                except StopIteration:
                    self._runningIterator = None
                    composeDone = True

        # If there is a compose block and it has finished, we're done
        if self._compose is not None and composeDone:
            return self._stop('finished compose block')

        # Optionally end when all our agents' behaviors have ended
        if self._endWithBehaviors:
            if all(agent.behavior._isFinished for agent in self._agents):
                return self._stop('all behaviors finished')

        # Check if any termination conditions apply
        for req in self._terminationConditions:
            if req.isTrue():
                return self._stop(req)

        # Scenario will not terminate yet
        return None

    def _stop(self, reason):
        veneer.endScenario(self, reason)
        self._runningIterator = None
        return reason

    def _invokeInner(self, agent, subs):
        for sub in subs:
            if not isinstance(sub, DynamicScenario):
                raise RuntimeParseError(f'expected a scenario, got {sub}')
            sub._prepare()
            sub._start()
        self._subScenarios = list(subs)
        while True:
            newSubs = []
            for sub in self._subScenarios:
                terminationReason = sub._step()
                if isinstance(terminationReason, EndSimulationAction):
                    yield terminationReason
                elif terminationReason is None:
                    newSubs.append(sub)
            self._subScenarios = newSubs
            if not newSubs:
                return
            yield None

    def _checkAlwaysRequirements(self):
        for req in self._alwaysRequirements:
            if not req.isTrue():
                # always requirements should never be violated at time 0, since
                # they are enforced during scene sampling
                assert veneer.currentSimulation.currentTime > 0
                raise RejectSimulationException(str(req))
        for sub in self._subScenarios:
            sub._checkAlwaysRequirements()

    def _runMonitors(self):
        terminationReason = None
        for monitor in self._monitors:
            action = monitor.step()
            if isinstance(action, EndSimulationAction):
                terminationReason = action
                # do not exit early, since subsequent monitors could reject the simulation
        for sub in self._subScenarios:
            subreason = sub._runMonitors()
            if subreason is not None:
                terminationReason = subreason
        return terminationReason

    def _checkSimulationTerminationConditions(self):
        for req in self._terminateSimulationConditions:
            if req.isTrue():
                return req
        return None

    @property
    def _allAgents(self):
        agents = list(self._agents)
        for sub in self._subScenarios:
            agents.extend(sub._allAgents)
        return agents

    def _inherit(self, other):
        self._objects.extend(other._objects)
        self._agents.extend(other._agents)
        self._globalParameters.update(other._globalParameters)
        self._externalParameters.extend(other._externalParameters)
        self._requirements.extend(other._requirements)
        self._behaviors.extend(other._behaviors)

    def _registerObject(self, obj):
        self._objects.append(obj)
        if getattr(obj, 'behavior', None) is not None:
            self._agents.append(obj)

    def _addRequirement(self, ty, reqID, req, line, prob):
        """Save a requirement defined at compile-time for later processing."""
        assert reqID not in self._pendingRequirements
        preq = PendingRequirement(ty, req, line, prob, self._ego)
        self._pendingRequirements[reqID] = preq

    def _addDynamicRequirement(self, ty, req, line):
        """Add a requirement defined during a dynamic simulation."""
        assert ty is not RequirementType.require
        dreq = DynamicRequirement(ty, req, line)
        self._registerCompiledRequirement(dreq)

    def _compileRequirements(self):
        namespace = self._dummyNamespace if self._dummyNamespace else self.__dict__
        requirementSyntax = self._requirementSyntax
        assert requirementSyntax is not None
        for reqID, requirement in self._pendingRequirements.items():
            syntax = requirementSyntax[reqID] if requirementSyntax else None
            compiledReq = requirement.compile(namespace, self, syntax)

            self._registerCompiledRequirement(compiledReq)
            self._requirementDeps.update(compiledReq.dependencies)

    def _registerCompiledRequirement(self, req):
        if req.ty is RequirementType.require:
            place = self._requirements
        elif req.ty is RequirementType.requireAlways:
            place = self._alwaysRequirements
        elif req.ty is RequirementType.terminateWhen:
            place = self._terminationConditions
        elif req.ty is RequirementType.terminateSimulationWhen:
            place = self._terminateSimulationConditions
        else:
            raise RuntimeError(f'internal error: requirement {req} has unknown type!')
        place.append(req)

    def _setTimeLimit(self, timeLimit, inSeconds=True):
        self._timeLimit = timeLimit
        self._timeLimitIsInSeconds = inSeconds

    def _toScenario(self, namespace):
        assert self._prepared

        if self._ego is None and self._compose is None:
            msg = 'did not specify ego object'
            modScenarios = namespace['_scenarios']
            if self._dummyNamespace and len(modScenarios) == 1:
                if modScenarios[0]._requiresArguments():
                    msg += ('\n(Note: this Scenic file contains a modular scenario, but it\n'
                            'cannot be used as the top-level scenario since it requires\n'
                            'arguments; so the whole file is being used as a top-level\n'
                            'scenario and needs an ego object.)')
                else:
                    msg += ('\n(Note: this Scenic file contains a modular scenario, but also\n'
                            'other code; so the whole file is being used as a top-level\n'
                            'scenario and needs an ego object.)')
            raise InvalidScenarioError(msg)

        # Extract workspace, if one is specified
        if 'workspace' in namespace:
            workspace = namespace['workspace']
            if not isinstance(workspace, Workspace):
                raise InvalidScenarioError(f'workspace {workspace} is not a Workspace')
            if needsSampling(workspace):
                raise InvalidScenarioError('workspace must be a fixed region')
            if needsLazyEvaluation(workspace):
                raise InvalidScenarioError('workspace uses value undefined '
                                           'outside of object definition')
        else:
            workspace = None

        from scenic.core.scenarios import Scenario
        scenario = Scenario(workspace, self._simulatorFactory,
                            self._objects, self._ego,
                            self._globalParameters, self._externalParameters,
                            self._requirements, self._requirementDeps,
                            self._monitors, self._behaviorNamespaces,
                            self)   # TODO unify these!
        return scenario

    def __str__(self):
        if self._dummyNamespace:
            return 'top-level scenario'
        else:
            args = argsToString(itertools.chain(self._args, self._kwargs.items()))
            return self.__class__.__name__ + args

# Behaviors

class Behavior(Invocable, Samplable):
    def __init_subclass__(cls):
        if cls.__module__ is not __name__:
            veneer.currentScenario._behaviors.append(cls)

    def __init__(self, *args, **kwargs):
        self.args = tuple(toDistribution(arg) for arg in args)
        self.kwargs = { name: toDistribution(arg) for name, arg in kwargs.items() }
        if not inspect.isgeneratorfunction(self.makeGenerator):
            raise RuntimeParseError(f'{self} does not take any actions'
                                    ' (perhaps you forgot to use "take" or "do"?)')

        # Validate arguments to the behavior
        sig = inspect.signature(self.makeGenerator)
        try:
            sig.bind(None, *args, **kwargs)
        except TypeError as e:
            raise RuntimeParseError(str(e)) from e
        Samplable.__init__(self, itertools.chain(self.args, self.kwargs.values()))
        Invocable.__init__(self)

    def sampleGiven(self, value):
        args = (value[arg] for arg in self.args)
        kwargs = { name: value[val] for name, val in self.kwargs.items() }
        return type(self)(*args, **kwargs)

    def start(self, agent):
        self.agent = agent
        self._runningIterator = self.makeGenerator(agent, *self.args, **self.kwargs)

    def step(self):
        assert self._runningIterator
        with veneer.executeInBehavior(self):
            try:
                actions = self._runningIterator.send(None)
            except StopIteration:
                actions = ()    # behavior ended early
        return actions

    def stop(self):
        self.agent = None
        self._runningIterator = None

    @property
    def _isFinished(self):
        return self._runningIterator is None

    def _invokeInner(self, agent, subs):
        assert len(subs) == 1
        sub = subs[0]
        if not isinstance(sub, Behavior):
            raise RuntimeParseError(f'expected a behavior, got {sub}')
        sub.start(agent)
        with veneer.executeInBehavior(sub):
            try:
                yield from sub._runningIterator
            finally:
                sub.stop()

    def __str__(self):
        args = argsToString(itertools.chain(self.args, self.kwargs.items()))
        return self.__class__.__name__ + args

def makeTerminationAction(line):
    assert not veneer.isActive()
    return EndSimulationAction(line)

# Monitors

class Monitor(Behavior):
    def __init_subclass__(cls):
        super().__init_subclass__()
        veneer.currentScenario._monitors.append(cls())

    def start(self):
        return super().start(None)

monitorPrefix = '_Scenic_monitor_'
def functionForMonitor(name):
    return monitorPrefix + name
def isAMonitorName(name):
    return name.startswith(monitorPrefix)
def monitorName(name):
    return name[len(monitorPrefix):]

# Guards

class GuardViolation(Exception):
    """Abstract exception raised when a guard of a behavior is violated.

    This will never be raised directly; either of the subclasses `PreconditionViolation`
    or `InvariantViolation` will be used, as appropriate.
    """
    violationType = 'guard'

    def __init__(self, behavior, lineno):
        self.behaviorName = behavior.__class__.__name__
        self.lineno = lineno

    def __str__(self):
        return f'violated {self.violationType} of {self.behaviorName} on line {self.lineno}'

class PreconditionViolation(GuardViolation):
    """Raised when a precondition is violated when invoking a behavior."""
    violationType = 'precondition'

class InvariantViolation(GuardViolation):
    """Raised when an invariant is violated when invoking/resuming a behavior."""
    violationType = 'invariant'

# Try-interrupt blocks

def runTryInterrupt(behavior, agent, body, conditions, handlers):
    body = InterruptBlock(None, body)
    interrupts = [InterruptBlock(c, h) for c, h in zip(conditions, handlers)]
    while True:
        # find next block to run, if any
        block = body
        for interrupt in interrupts:
            if interrupt.isEnabled or interrupt.isRunning:
                block = interrupt
                break
        result, concluded = block.step(behavior, agent)
        if concluded:
            if result is BlockConclusion.FINISHED and block is not body:
                continue    # interrupt handler finished
            else:
                return result   # entire try-interrupt statement will terminate
        else:
            yield result
            behavior.checkInvariants()

@enum.unique
class BlockConclusion(enum.Enum):
    FINISHED = enum.auto()
    ABORT = enum.auto()
    RETURN = enum.auto()
    BREAK = enum.auto()
    CONTINUE = enum.auto()

    def __call__(self, value):
        self.return_value = value
        return self

class InterruptBlock:
    def __init__(self, condition, body):
        self.condition = condition
        self.body = body
        self.runningIterator = None

    @property
    def isEnabled(self):
        with veneer.executeInGuard():
            return bool(self.condition())

    @property
    def isRunning(self):
        return self.runningIterator is not None

    def step(self, behavior, agent):
        if not self.runningIterator:
            it = self.body(behavior, agent)
            if not isinstance(it, types.GeneratorType):
                return (it, True)
            self.runningIterator = it
        try:
            result = self.runningIterator.send(None)
            return (result, False)
        except StopIteration as e:
            self.runningIterator = None
            return (e.value, True)

import scenic.syntax.veneer as veneer




import pickle

import scenic.simulators as simulators
import scenic.simulators.webots.utils as utils
from scenic.core.vectors import Vector

class WebotsSimulator(simulators.Simulator):
    def __init__(self, supervisor):
        super().__init__()
        self.supervisor = supervisor

    def createSimulation(self, scene):
        return WebotsSimulation(scene, self.supervisor)

class WebotsSimulation(simulators.Simulation):
    def __init__(self, scene, supervisor):
        super().__init__(scene)
        self.supervisor = supervisor
        self.timeStep = int(supervisor.getBasicTimeStep())
        self.objects = scene.objects

        # Find Webots objects corresponding to Scenic objects
        self.webotsObjects = {}
        for obj in self.objects:
            if not hasattr(obj, 'webotsName'):
                raise RuntimeError(f'object {obj} does not have a webotsName property')
            name = obj.webotsName
            webotsObj = supervisor.getFromDef(name)
            if webotsObj is None:
                raise RuntimeError(f'Webots object {name} does not exist in world')
            self.webotsObjects[obj] = webotsObj
            obj.webotsObject = webotsObj

            # get starting elevation
            pos = webotsObj.getField('translation').getSFVec3f()
            obj.elevation = pos[1]

        # Reset Webots and object controllers
        supervisor.simulationResetPhysics()
        for webotsObj in self.webotsObjects.values():
            webotsObj.restartController()

        # Set initial positions and orientations of Webots objects
        self.writePropertiesToWebots()

    def writePropertiesToWebots(self):
        for obj in self.objects:
            webotsObj = self.webotsObjects[obj]
            # position
            pos = utils.scenicToWebotsPosition(obj.position, y=obj.elevation)
            webotsObj.getField('translation').setSFVec3f(pos)
            # heading
            rot = utils.scenicToWebotsRotation(obj.heading)
            webotsObj.getField('rotation').setSFRotation(rot)

    def readPropertiesFromWebots(self):
        for obj in self.objects:
            webotsObj = self.webotsObjects[obj]
            # webotsObject
            obj.webotsObject = webotsObj
            # position
            pos = webotsObj.getField('translation').getSFVec3f()
            x, y = utils.webotsToScenicPosition(pos)
            obj.position = Vector(x, y)
            # heading
            rot = webotsObj.getField('rotation').getSFRotation()
            heading = utils.webotsToScenicRotation(rot, tolerance2D=100)
            if heading is None:
                raise RuntimeError(f'{webotsObj} has non-planar orientation!')
            obj.heading = heading

    def currentState(self):
        return tuple(obj.position for obj in self.objects)

    def initialState(self):
        return self.currentState()

    def step(self, actions):
        # execute actions
        for agent, action in actions.items():
            if action is not None:
                action.applyTo(agent, self.webotsObjects[agent])
        # run simulation for one time step
        self.supervisor.step(self.timeStep)
        # read back the results of the simulation
        self.readPropertiesFromWebots()
        return self.currentState()

class MoveAction(simulators.Action):
    def __init__(self, offset):
        self.offset = offset

    def applyTo(self, obj, webotsObj):
        pos = obj.position.offsetRotated(obj.heading, self.offset)
        pos = utils.scenicToWebotsPosition(pos, y=obj.elevation)
        webotsObj.getField('translation').setSFVec3f(pos)

class WriteFileAction(simulators.Action):
    def __init__(self, path, data):
        self.path = path
        self.data = data

    def applyTo(self, obj, webotsObj):
        with open(self.path, 'wb') as outFile:
            pickle.dump(self.data, outFile)



# Copyright (c) # Copyright (c) 2018-2020 CVC.
#
# This work is licensed under the terms of the MIT license.
# For a copy, see <https://opensource.org/licenses/MIT>.

""" This module contains PID controllers to perform lateral and longitudinal control. """

from collections import deque
import math
import numpy as np
import carla
from scenic.simulators.carla.misc import get_speed


class VehiclePIDController():
    """
    VehiclePIDController is the combination of two PID controllers
    (lateral and longitudinal) to perform the
    low level control a vehicle from client side
    """


    def __init__(self, vehicle, args_lateral=None, args_longitudinal=None, max_throttle=0.75, max_brake=0.3, max_steering=0.8):
        """
        Constructor method.

        :param vehicle: actor to apply to local planner logic onto
        :param args_lateral: dictionary of arguments to set the lateral PID controller
        using the following semantics:
            K_P -- Proportional term
            K_D -- Differential term
            K_I -- Integral term
        :param args_longitudinal: dictionary of arguments to set the longitudinal
        PID controller using the following semantics:
            K_P -- Proportional term
            K_D -- Differential term
            K_I -- Integral term
        """

        self.max_brake = max_brake
        self.max_throt = max_throttle
        self.max_steer = max_steering

        self._vehicle = vehicle
        self._world = self._vehicle.get_world()
        self.past_steering = self._vehicle.get_control().steer

        if args_longitudinal!=None:
            self._lon_controller = PIDLongitudinalController(self._vehicle, **args_longitudinal)
        else:
            self._lon_controller = PIDLongitudinalController(self._vehicle)

        if args_lateral!=None:
            self._lat_controller = PIDLateralController(self._vehicle, **args_lateral)
        else: 
            self._lat_controller = PIDLateralController(self._vehicle)

    def run_step(self, target_speed, waypoint):
        """
        Execute one step of control invoking both lateral and longitudinal
        PID controllers to reach a target waypoint
        at a given target_speed.

        :param target_speed: desired vehicle speed
        :param waypoint: target location encoded as a waypoint
        :return: distance (in meters) to the waypoint
        """

        acceleration = self._lon_controller.run_step(target_speed)
        current_steering = self._lat_controller.run_step(waypoint)
        control = carla.VehicleControl()
        if acceleration >= 0.0:
            control.throttle = min(acceleration, self.max_throt)
            control.brake = 0.0
        else:
            control.throttle = 0.0
            control.brake = min(abs(acceleration), self.max_brake)

        # Steering regulation: changes cannot happen abruptly, can't steer too much.

        if current_steering > self.past_steering + 0.1:
            current_steering = self.past_steering + 0.1
        elif current_steering < self.past_steering - 0.1:
            current_steering = self.past_steering - 0.1

        if current_steering >= 0:
            steering = min(self.max_steer, current_steering)
        else:
            steering = max(-self.max_steer, current_steering)

        control.steer = steering
        control.hand_brake = False
        control.manual_gear_shift = False
        self.past_steering = steering

        return control


class PIDLongitudinalController():
    """
    PIDLongitudinalController implements longitudinal control using a PID.
    """


    def __init__(self, vehicle, K_P=1.0, K_D=0.0, K_I=0.0, dt=0.03):
        """
        Constructor method.

            :param vehicle: actor to apply to local planner logic onto
            :param K_P: Proportional term
            :param K_D: Differential term
            :param K_I: Integral term
            :param dt: time differential in seconds
        """
        self._vehicle = vehicle
        self._k_p = K_P
        self._k_d = K_D
        self._k_i = K_I
        self._dt = dt
        self._error_buffer = deque(maxlen=10)

    def run_step(self, target_speed, debug=False):
        """
        Execute one step of longitudinal control to reach a given target speed.

            :param target_speed: target speed in Km/h
            :param debug: boolean for debugging
            :return: throttle control
        """
        current_speed = get_speed(self._vehicle)

        if debug:
            print('Current speed = {}'.format(current_speed))

        return self._pid_control(target_speed, current_speed)

    def _pid_control(self, target_speed, current_speed):
        """
        Estimate the throttle/brake of the vehicle based on the PID equations

            :param target_speed:  target speed in Km/h
            :param current_speed: current speed of the vehicle in Km/h
            :return: throttle/brake control
        """

        error = target_speed - current_speed
        self._error_buffer.append(error)

        if len(self._error_buffer) >= 2:
            _de = (self._error_buffer[-1] - self._error_buffer[-2]) / self._dt
            _ie = sum(self._error_buffer) * self._dt
        else:
            _de = 0.0
            _ie = 0.0

        return np.clip((self._k_p * error) + (self._k_d * _de) + (self._k_i * _ie), -1.0, 1.0)

class PIDLateralController():
    """
    PIDLateralController implements lateral control using a PID.
    """

    def __init__(self, vehicle, K_P=1.0, K_D=0.0, K_I=0.0, dt=0.03):
        """
        Constructor method.

            :param vehicle: actor to apply to local planner logic onto
            :param K_P: Proportional term
            :param K_D: Differential term
            :param K_I: Integral term
            :param dt: time differential in seconds
        """
        self._vehicle = vehicle
        self._k_p = K_P
        self._k_d = K_D
        self._k_i = K_I
        self._dt = dt
        self._e_buffer = deque(maxlen=10)

    def run_step(self, waypoint):
        """
        Execute one step of lateral control to steer
        the vehicle towards a certain waypoin.

            :param waypoint: target waypoint
            :return: steering control in the range [-1, 1] where:
                -1 maximum steering to left
                +1 maximum steering to right
        """
        return self._pid_control(waypoint, self._vehicle.get_transform())

    def _pid_control(self, waypoint, vehicle_transform):
        """
        Estimate the steering angle of the vehicle based on the PID equations

            :param waypoint: target waypoint
            :param vehicle_transform: current transform of the vehicle
            :return: steering control in the range [-1, 1]
        """
        v_begin = vehicle_transform.location
        v_end = v_begin + carla.Location(x=math.cos(math.radians(vehicle_transform.rotation.yaw)),
                                         y=math.sin(math.radians(vehicle_transform.rotation.yaw)))

        v_vec = np.array([v_end.x - v_begin.x, v_end.y - v_begin.y, 0.0])
        w_vec = np.array([waypoint.transform.location.x -
                          v_begin.x, waypoint.transform.location.y -
                          v_begin.y, 0.0])
        _dot = math.acos(np.clip(np.dot(w_vec, v_vec) /
                                 (np.linalg.norm(w_vec) * np.linalg.norm(v_vec)), -1.0, 1.0))

        _cross = np.cross(v_vec, w_vec)

        if _cross[2] < 0:
            _dot *= -1.0

        self._e_buffer.append(_dot)
        if len(self._e_buffer) >= 2:
            _de = (self._e_buffer[-1] - self._e_buffer[-2]) / self._dt
            _ie = sum(self._e_buffer) * self._dt
        else:
            _de = 0.0
            _ie = 0.0

        return np.clip((self._k_p * _dot) + (self._k_d * _de) + (self._k_i * _ie), -1.0, 1.0)


"""Simulator interface for LGSVL."""

import math
import warnings

import lgsvl

import scenic.core.simulators as simulators
import scenic.simulators.lgsvl.utils as utils
from scenic.syntax.veneer import verbosePrint
from scenic.core.vectors import Vector

class LGSVLSimulator(simulators.Simulator):
    def __init__(self, lgsvl_scene, address='localhost', port=8181, alwaysReload=False):
        super().__init__()
        verbosePrint('Connecting to LGSVL Simulator...')
        self.client = lgsvl.Simulator(address=address, port=port)
        if alwaysReload or self.client.current_scene != lgsvl_scene:
            self.client.load(scene=lgsvl_scene)
        verbosePrint('Map loaded in simulator.')

    def createSimulation(self, scene, verbosity=0):
        return LGSVLSimulation(scene, self.client, verbosity=verbosity)


class LGSVLSimulation(simulators.Simulation):
    def __init__(self, scene, client, verbosity=0):
        timestep = scene.params.get('time_step', 1.0/10)
        super().__init__(scene, timestep=timestep, verbosity=verbosity)
        self.client = client
        self.usingApollo = False
        self.data = {}
        self.collisionOccurred = False

        # Reset simulator (deletes all existing objects)
        self.client.reset()

        # Create LGSVL objects corresponding to Scenic objects
        for obj in self.objects:
            if not hasattr(obj, 'lgsvlObject'):
                continue    # not an LGSVL object
            self.createObjectInSimulator(obj)

    def createObjectInSimulator(self, obj):
        # Figure out what type of LGSVL object this is
        if not hasattr(obj, 'lgsvlName'):
            raise RuntimeError(f'object {obj} does not have an lgsvlName property')
        if not hasattr(obj, 'lgsvlAgentType'):
            raise RuntimeError(f'object {obj} does not have an lgsvlAgentType property')
        name = obj.lgsvlName
        agentType = obj.lgsvlAgentType

        # Set up position and orientation
        state = lgsvl.AgentState()
        elevation = obj.elevation
        if elevation is None:
            elevation = self.groundElevationAt(obj.position)
        state.transform.position = utils.scenicToLGSVLPosition(obj.position, elevation)
        state.transform.rotation = utils.scenicToLGSVLRotation(obj.heading)

        # Create LGSVL object
        lgsvlObj = self.client.add_agent(name, agentType, state)
        obj.lgsvlObject = lgsvlObj

        # Initialize Data
        self.data[obj] = {}
        # Initialize Apollo if needed
        if getattr(obj, 'apolloVehicle', None):
            self.initApolloFor(obj, lgsvlObj)

    def groundElevationAt(self, pos):
        origin = utils.scenicToLGSVLPosition(pos, 100000)
        result = self.client.raycast(origin, lgsvl.Vector(0, -1, 0), 1)
        if result is None:
            warnings.warn(f'no ground at position {pos}')
            return 0
        return result.point.y

    def initApolloFor(self, obj, lgsvlObj):
        """Initialize Apollo for an ego vehicle.

        Uses LG's interface which injects packets into Dreamview.
        """
        if self.usingApollo:
            raise RuntimeError('can only use one Apollo vehicle')
        self.usingApollo = True

        def on_collision(agent1, agent2, contact):
            if agent1 is not None and agent1.name == lgsvlObj.name:
                self.data[obj]['collision'] = True
            if agent2 is not None and agent2.name == lgsvlObj.name:
                self.data[obj]['collision'] = True
            if self.data[obj]['collision']:
                self.collisionOccurred = True

        # Initialize Data
        self.data[obj]['collision'] = False
        lgsvlObj.on_collision(on_collision)

        # connect bridge from LGSVL to Apollo
        lgsvlObj.connect_bridge(obj.bridgeHost, obj.bridgePort)

        # set up connection and map/vehicle configuration
        import dreamview
        dv = dreamview.Connection(self.client, lgsvlObj)
        obj.dreamview = dv
        waitToStabilize = False
        hdMap = self.scene.params['apolloHDMap']
        if dv.getCurrentMap() != hdMap:
            dv.setHDMap(hdMap)
            waitToStabilize = True
        if dv.getCurrentVehicle() != obj.apolloVehicle:
            dv.setVehicle(obj.apolloVehicle)
            waitToStabilize = True
        
        verbosePrint('Initializing Apollo...')

        # stop the car to cancel buffered speed from previous simulations
        cntrl = lgsvl.VehicleControl()
        cntrl.throttle = 0.0
        lgsvlObj.apply_control(cntrl, True)
        # start modules
        dv.disableModule('Control')
        ready = dv.getModuleStatus()
        for module in obj.apolloModules:
            if not ready[module]:
                dv.enableModule(module)
                verbosePrint(f'Module {module} is not ready...')
                waitToStabilize = True
        while True:
            ready = dv.getModuleStatus()
            if all(ready[module] for module in obj.apolloModules):
                break

        # wait for Apollo to stabilize, if needed
        if waitToStabilize:
            verbosePrint('Waiting for Apollo to stabilize...')
            self.client.run(25)
        dv.enableModule('Control')
        self.client.run(15)
        verbosePrint('Initialized Apollo.')

    def executeActions(self, allActions):
        super().executeActions(allActions)

        # Apply state/control updates which were accumulated while executing the actions
        for obj in self.agents:
            if obj._stateUpdated:
                obj.lgsvlObject.state = obj.state
                obj._stateUpdated = False
            ctrl = getattr(obj, '_control', None)
            if ctrl is not None:
                obj.lgsvlObject.apply_control(ctrl, obj._stickyControl)
                obj._control = None

    def step(self):
        self.client.run(time_limit=self.timestep)

    def getProperties(self, obj, properties):
        lgsvlObj = obj.lgsvlObject
        state = lgsvlObj.state
        obj.state = state   # cache state for subsequent updates

        velocity = utils.lgsvlToScenicPosition(state.velocity)
        speed = math.hypot(*velocity)

        values = dict(
            position=utils.lgsvlToScenicPosition(state.position),
            elevation=utils.lgsvlToScenicElevation(state.position),
            heading=utils.lgsvlToScenicRotation(state.rotation),
            velocity=velocity,
            speed=speed,
            angularSpeed=utils.lgsvlToScenicAngularSpeed(state.rotation),
        )
        return values



"""A compiler and scene generator for the Scenic scenario description language.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree:

   core
   domains
   formats
   simulators
   syntax
"""

from .syntax.translator import scenarioFromFile, scenarioFromString

import scenic.core.errors as _errors
_errors.showInternalBacktrace = False
del _errors




"""Abstract interface to simulators supporting the driving domain."""

from scenic.core.simulators import Simulator, Simulation

class DrivingSimulator(Simulator):
    def createSimulation(self, scene):
        raise NotImplementedError

class DrivingSimulation(Simulation):
    pass



"""Utility functions for geometric computation."""

import math
import itertools
import warnings

import numpy as np
import shapely.geometry
import shapely.ops

from scenic.core.distributions import (needsSampling, distributionFunction,
                                       monotonicDistributionFunction)
from scenic.core.lazy_eval import needsLazyEvaluation
from scenic.core.utils import cached_property

@distributionFunction
def sin(x) -> float:
	return math.sin(x)

@distributionFunction
def cos(x) -> float:
	return math.cos(x)

@monotonicDistributionFunction
def hypot(x, y) -> float:
	return math.hypot(x, y)

@monotonicDistributionFunction
def max(*args, **kwargs):
	return __builtins__['max'](*args, **kwargs)

@monotonicDistributionFunction
def min(*args, **kwargs):
	return __builtins__['min'](*args, **kwargs)

@distributionFunction
def normalizeAngle(angle) -> float:
	while angle > math.pi:
		angle -= math.tau
	while angle < -math.pi:
		angle += math.tau
	assert -math.pi <= angle <= math.pi
	return angle

def addVectors(a, b):
	ax, ay = a[0], a[1]
	bx, by = b[0], b[1]
	return (ax + bx, ay + by)

def subtractVectors(a, b):
	ax, ay = a[0], a[1]
	bx, by = b[0], b[1]
	return (ax - bx, ay - by)

def averageVectors(a, b, weight=0.5):
	ax, ay = a[0], a[1]
	bx, by = b[0], b[1]
	aw, bw = 1.0 - weight, weight
	return (ax * aw + bx * bw, ay * aw + by * bw)

def rotateVector(vector, angle):
	x, y = vector
	c, s = cos(angle), sin(angle)
	return ((c * x) - (s * y), (s * x) + (c * y))

def findMinMax(iterable):
	minv = float('inf')
	maxv = float('-inf')
	for val in iterable:
		if val < minv:
			minv = val
		if val > maxv:
			maxv = val
	return (minv, maxv)

def radialToCartesian(point, radius, heading):
	angle = heading + (math.pi / 2.0)
	rx, ry = radius * cos(angle), radius * sin(angle)
	return (point[0] + rx, point[1] + ry)

def positionRelativeToPoint(point, heading, offset):
	ro = rotateVector(offset, heading)
	return addVectors(point, ro)

def headingOfSegment(pointA, pointB):
	ax, ay = pointA
	bx, by = pointB
	return normalizeAngle(math.atan2(by - ay, bx - ax) - (math.pi / 2.0))

def viewAngleToPoint(point, base, heading):
	x, y = base
	ox, oy = point
	a = math.atan2(oy - y, ox - x) - (heading + (math.pi / 2.0))
	return normalizeAngle(a)

def apparentHeadingAtPoint(point, heading, base):
	x, y = base
	ox, oy = point
	a = (heading + (math.pi / 2.0)) - math.atan2(oy - y, ox - x)
	return normalizeAngle(a)

def circumcircleOfAnnulus(center, heading, angle, minDist, maxDist):
	m = (minDist + maxDist) / 2.0
	g = (maxDist - minDist) / 2.0
	h = m * math.sin(angle / 2.0)
	h2 = h * h
	d = math.sqrt(h2 + (m * m))
	r = math.sqrt(h2 + (g * g))
	return radialToCartesian(center, d, heading), r

def pointIsInCone(point, base, heading, angle):
	va = viewAngleToPoint(point, base, heading)
	return (abs(va) <= angle / 2.0)

def distanceToLine(point, a, b):
	lx, ly = b[0] - a[0], b[1] - a[1]
	norm = math.hypot(lx, ly)
	nx, ny = -ly/norm, lx/norm
	px, py = point[0] - a[0], point[1] - a[1]
	return abs((px * nx) + (py * ny))

def distanceToSegment(point, a, b):
	lx, ly = b[0] - a[0], b[1] - a[1]
	px, py = point[0] - a[0], point[1] - a[1]
	proj = (px * lx) + (py * ly)
	if proj < 0:
		return math.hypot(px, py)
	lnorm = math.hypot(lx, ly)
	if proj > lnorm * lnorm:
		return math.hypot(px - lx, py - ly)
	else:
		return abs((py * lx) - (px * ly)) / lnorm

def polygonUnion(polys, buf=0, tolerance=0, holeTolerance=0.002):
	if not polys:
		return shapely.geometry.Polygon()
	polys = list(polys)
	if len(polys) == 1:
		return polys[0]
	buffered = [poly.buffer(buf) for poly in polys]
	union = shapely.ops.unary_union(buffered).buffer(-buf)
	assert union.is_valid, union
	if tolerance > 0:
		union = cleanPolygon(union, tolerance, holeTolerance)
		assert union.is_valid, union
		#checkPolygon(union, tolerance)
	return union

def checkPolygon(poly, tolerance):
	def checkPolyline(pl):
		for i, p in enumerate(pl[:-1]):
			q = pl[i+1]
			dx, dy = q[0] - p[0], q[1] - p[1]
			assert math.hypot(dx, dy) >= tolerance
	if isinstance(poly, shapely.geometry.MultiPolygon):
		for p in poly:
			checkPolygon(p, tolerance)
	else:
		checkPolyline(poly.exterior.coords)
		for i in poly.interiors:
			checkPolyline(i.coords)

def cleanPolygon(poly, tolerance, holeTolerance=0, minRelArea=0.05, minHullLenRatio=0.9):
	if holeTolerance == 0:
		holeTolerance = tolerance * tolerance
	if poly.is_empty:
		return poly
	elif isinstance(poly, shapely.geometry.MultiPolygon):
		polys = [cleanPolygon(p, tolerance, holeTolerance) for p in poly]
		total = sum(poly.area for poly in polys)
		kept = []
		for poly in polys:
			area = poly.area
			if area >= holeTolerance or area >= minRelArea * total:
				kept.append(poly)
		poly = shapely.ops.unary_union(kept)
		assert poly.is_valid, poly
		return poly
	exterior = poly.exterior.simplify(tolerance)
	newExterior = cleanChain(exterior.coords, tolerance)
	if len(newExterior) <= 3:
		# attempt to save very thin polygons that would get reduced to a single point
		hull = exterior.convex_hull
		if hull.length >= minHullLenRatio * exterior.length:
			return hull
		return shapely.geometry.Polygon()
	ints = []
	for interior in poly.interiors:
		interior = interior.simplify(tolerance)
		interior = cleanChain(interior.coords, tolerance)
		if len(interior) >= 4:
			hole = shapely.geometry.Polygon(interior)
			if hole.area > holeTolerance:
				ints.append(interior)
	newPoly = shapely.geometry.Polygon(newExterior, ints)
	if not newPoly.is_valid:
		# last-ditch attempt to salvage polygon by splitting across self-intersections
		ext = splitSelfIntersections(newPoly.exterior, minArea=holeTolerance,
		                             minRelArea=minRelArea, minHullLenRatio=minHullLenRatio)
		holes = [
			splitSelfIntersections(hole, minArea=holeTolerance,
			                       minRelArea=minRelArea, minHullLenRatio=minHullLenRatio)
			for hole in newPoly.interiors
		]
		newPoly = ext.difference(shapely.ops.unary_union(holes))
	assert newPoly.is_valid, newPoly
	return newPoly

def splitSelfIntersections(chain, minArea, minRelArea, minHullLenRatio):
	ls = shapely.geometry.LineString(chain)
	parts = list(shapely.ops.polygonize(shapely.ops.unary_union(ls)))
	total = sum(part.area for part in parts)
	if total < minArea:
		# attempt to save very thin polygons that could get thrown out by polygonize
		hull = ls.convex_hull
		if hull.length >= minHullLenRatio * ls.length:
			return hull
	kept = [part for part in parts if part.area >= minArea or (part.area / total) >= minRelArea]
	return shapely.ops.unary_union(kept)

def cleanChain(chain, tolerance=1e-6, lineTolerance=1e-6):
	closed = (tuple(chain[0]) == tuple(chain[-1]))
	minLength = 4 if closed else 3
	if len(chain) <= minLength:
		return chain
	tol2 = tolerance * tolerance
	# collapse nearby points (since Shapely's simplify method doesn't always do this)
	a = chain[0]
	newChain = [a]
	for b in chain[1:]:
		dx, dy = b[0] - a[0], b[1] - a[1]
		if (dx * dx) + (dy * dy) > tol2:
			newChain.append(b)
			a = b
	if closed:
		b = chain[0]
		dx, dy = b[0] - a[0], b[1] - a[1]
		if (dx * dx) + (dy * dy) <= tol2:
			newChain.pop()
		newChain.append(chain[0])
	if len(newChain) <= minLength:
		if not closed:
			if len(newChain) > 1:
				newChain.pop()
			newChain.append(chain[-1])
		return newChain
	# collapse collinear points
	chain = newChain
	if closed:
		a = chain[-2]
		b = chain[0]
		ci = 1
		newChain = []
	else:
		a = chain[0]
		b = chain[1]
		ci = 2
		newChain = [a]
	for c in chain[ci:]:
		dx, dy = c[0] - a[0], c[1] - a[1]
		if dx == dy == 0 or distanceToLine(b, a, c) > lineTolerance:
			newChain.append(b)
			a = b
			b = c
		else:
			b = c
	newChain.append(c)
	return newChain

def removeHoles(polygon):
	if polygon.is_empty:
		return polygon
	elif isinstance(polygon, shapely.geometry.MultiPolygon):
		polys = (removeHoles(poly) for poly in polygon)
		poly = shapely.geometry.MultiPolygon(polys)
		assert poly.is_valid, poly
		return poly
	else:
		return shapely.geometry.Polygon(polygon.exterior)

class TriangulationError(RuntimeError):
	"""Signals that the installed triangulation libraries are insufficient.

	Specifically, raised when pypoly2tri hits the recursion limit trying to
	triangulate a large polygon.
	"""
	pass

def triangulatePolygon(polygon):
	"""Triangulate the given Shapely polygon.

	Note that we can't use ``shapely.ops.triangulate`` since it triangulates
	point sets, not polygons (i.e., it doesn't respect edges). We need an
	algorithm for triangulation of polygons with holes (it doesn't need to be a
	Delaunay triangulation).

	We use ``mapbox_earcut`` by default. If it is not installed, we allow fallback to
	``pypoly2tri`` for historical reasons (we originally used the GPC library, which is
	not free for commercial use, falling back to ``pypoly2tri`` if not installed).

	Args:
		polygon (shapely.geometry.Polygon): Polygon to triangulate.

	Returns:
		A list of disjoint (except for edges) triangles whose union is the
		original polygon.
	"""
	try:
		return triangulatePolygon_mapbox(polygon)
	except ImportError:
		pass
	try:
		return triangulatePolygon_pypoly2tri(polygon)
	except ImportError:
		pass
	raise RuntimeError('no triangulation libraries installed '
	                   '(did you uninstall mapbox_earcut?)')

def triangulatePolygon_pypoly2tri(polygon):
	import pypoly2tri
	polyline = []
	for c in polygon.exterior.coords[:-1]:
		polyline.append(pypoly2tri.shapes.Point(c[0],c[1]))
	cdt = pypoly2tri.cdt.CDT(polyline)
	for i in polygon.interiors:
		polyline = []
		for c in i.coords[:-1]:
			polyline.append(pypoly2tri.shapes.Point(c[0],c[1]))
		cdt.AddHole(polyline)
	try:
		cdt.Triangulate()
	except RecursionError:		# polygon too big for pypoly2tri
		raise TriangulationError('pypoly2tri unable to triangulate large polygon; for '
		                         'non-commercial use, try "pip install Polygon3"')

	triangles = list()
	for t in cdt.GetTriangles():
		triangles.append(shapely.geometry.Polygon([
			t.GetPoint(0).toTuple(),
			t.GetPoint(1).toTuple(),
			t.GetPoint(2).toTuple()
		]))
	return triangles

def triangulatePolygon_mapbox(polygon):
	import mapbox_earcut
	vertices, rings = [], []
	ring = polygon.exterior.coords[:-1]		# despite 'ring' name, actually need a chain
	vertices.extend(ring)
	rings.append(len(vertices))
	for interior in polygon.interiors:
		ring = interior.coords[:-1]
		vertices.extend(ring)
		rings.append(len(vertices))
	vertices = np.array(vertices, dtype=np.float64)
	rings = np.array(rings)
	result = mapbox_earcut.triangulate_float64(vertices, rings)

	triangles = []
	points = vertices[result]
	its = [iter(points)] * 3
	for triple in zip(*its):
		triangles.append(shapely.geometry.Polygon(triple))
	return triangles

def plotPolygon(polygon, plt, style='r-', **kwargs):
	def plotCoords(chain):
		x, y = chain.xy
		plt.plot(x, y, style, **kwargs)
	if polygon.is_empty:
		return
	if isinstance(polygon, (shapely.geometry.MultiPolygon,
	                        shapely.geometry.MultiLineString,
	                        shapely.geometry.MultiPoint,
	                        shapely.geometry.collection.GeometryCollection)):
		polygons = polygon
	else:
		polygons = [polygon]
	for polygon in polygons:
		if isinstance(polygon, shapely.geometry.Polygon):
			plotCoords(polygon.exterior)
			for ring in polygon.interiors:
				plotCoords(ring)
		elif isinstance(polygon, (shapely.geometry.LineString, shapely.geometry.LinearRing,
		                          shapely.geometry.Point)):
			plotCoords(polygon)
		else:
			raise RuntimeError(f'unknown kind of shapely geometry {polygon}')

class _RotatedRectangle:
	"""mixin providing collision detection for rectangular objects and regions"""
	def containsPoint(self, point):
		pt = shapely.geometry.Point(point)
		return self.polygon.intersects(pt)

	def intersects(self, rect):
		return self.polygon.intersects(rect.polygon)

	@cached_property
	def polygon(self):
		position, heading, hw, hl = self.position, self.heading, self.hw, self.hl
		if any(needsSampling(c) or needsLazyEvaluation(c)
		       for c in (position, heading, hw, hl)):
			return None		# can only convert fixed Regions to Polygons
		corners = _RotatedRectangle.makeCorners(position.x, position.y, heading, hw, hl)
		return shapely.geometry.Polygon(corners)

	@staticmethod
	def makeCorners(px, py, heading, hw, hl):
		s, c = sin(heading), cos(heading)
		s_hw, c_hw = s*hw, c*hw
		s_hl, c_hl = s*hl, c*hl
		corners = (
			(px + c_hw - s_hl, py + s_hw + c_hl),
			(px - c_hw - s_hl, py - s_hw + c_hl),
			(px - c_hw + s_hl, py - s_hw - c_hl),
			(px + c_hw + s_hl, py + s_hw - c_hl)
		)
		return corners



"""Interface to the CARLA driving simulator.

This interface has been tested with `CARLA <https://carla.org/>`_ versions 0.9.9,
0.9.10, and 0.9.11.
It supports dynamic scenarios involving vehicles, pedestrians, and props.

The interface implements the :obj:`scenic.domains.driving` abstract domain, so any
object types, behaviors, utility functions, etc. from that domain may be used freely.
For details of additional CARLA-specific functionality, see the world model
:obj:`scenic.simulators.carla.model`.

.. raw:: html

   <h2>Submodules</h2>

.. autosummary::
   :toctree:

   model
   actions
   behaviors
   simulator
   blueprints
   controller
   misc
"""



"""Low-level controllers useful for vehicles.

The Lateral/Longitudinal PID controllers are adapted from `CARLA`_'s PID controllers,
which are licensed under the following terms:

	Copyright (c) 2018-2020 CVC.

	This work is licensed under the terms of the MIT license.
	For a copy, see <https://opensource.org/licenses/MIT>.


.. _CARLA: https://carla.org/
"""

from collections import deque

import numpy as np
import scipy.linalg as linalg

class PIDLongitudinalController:
	"""Longitudinal control using a PID to reach a target speed.

	Arguments:
		K_P: Proportional gain
		K_D: Derivative gain
		K_I: Integral gain
		dt: time step
	"""
	def __init__(self, K_P=0.5, K_D=0.1, K_I=0.2, dt=0.1):
		self._k_p = K_P
		self._k_d = K_D
		self._k_i = K_I
		self._dt = dt
		self._error_buffer = deque(maxlen=10)

	def run_step(self, speed_error):
		"""Estimate the throttle/brake of the vehicle based on the PID equations.

		Arguments:
			speed_error: target speed minus current speed

		Returns:
			a signal between -1 and 1, with negative values indicating braking.
		"""
		error = speed_error
		self._error_buffer.append(error)

		if len(self._error_buffer) >= 2:
			_de = (self._error_buffer[-1] - self._error_buffer[-2]) / self._dt
			_ie = sum(self._error_buffer) * self._dt
		else:
			_de = 0.0
			_ie = 0.0

		return np.clip((self._k_p * error) + (self._k_d * _de) + (self._k_i * _ie),
		               -1.0, 1.0)

class PIDLateralController:
	"""Lateral control using a PID to track a trajectory.

	Arguments:
		K_P: Proportional gain
		K_D: Derivative gain
		K_I: Integral gain
		dt: time step
	"""
	def __init__(self, K_P=0.3, K_D=0.2, K_I=0, dt=0.1):
		self.Kp = K_P
		self.Kd = K_D
		self.Ki = K_I
		self.PTerm = 0
		self.ITerm = 0
		self.DTerm = 0
		self.dt = dt
		self.last_error = 0
		self.windup_guard = 20.0
		self.output = 0

	def run_step(self, cte):
		"""Estimate the steering angle of the vehicle based on the PID equations.

		Arguments:
			cte: cross-track error (distance to right of desired trajectory)

		Returns:
			a signal between -1 and 1, with -1 meaning maximum steering to the left.
		"""
		error = cte
		delta_error = error - self.last_error
		self.PTerm = self.Kp * error
		self.ITerm += error * self.dt

		if (self.ITerm < -self.windup_guard):
			self.ITerm = -self.windup_guard
		elif (self.ITerm > self.windup_guard):
			self.ITerm = self.windup_guard

		self.DTerm = delta_error / self.dt

		# Remember last error for next calculation
		self.last_error = error

		self.output = self.PTerm + (self.Ki * self.ITerm) + (self.Kd * self.DTerm)

		return np.clip(self.output, -1, 1)

class LQR:
	def __init__(self,v_target, wheelbase, Q, R):
		self.v_target = v_target
		self.wheelbase = wheelbase
		self.Q = Q
		self.R = R

	def run_step(self):
		A = np.matrix([[0, self.v_target*(5./18.)], [0, 0]])
		B = np.matrix([[0], [(self.v_target/self.wheelbase)*(5./18.)]])
		V = np.matrix(linalg.solve_continuous_are(A, B, self.Q, self.R))
		K = np.matrix(linalg.inv(self.R)*(B.T*V))
		return K


# This file utilizes existing CARLA code:

# Copyright (c) 2018 Intel Labs.
# authors: German Ros (german.ros@intel.com)
#
# This work is licensed under the terms of the MIT license.
# Permission is hereby granted, free of charge, to any person
# obtaining a copy of this software and associated documentation
# files (the "Software"), to deal in the Software without
# restriction, including without limitation the rights to use,
# copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
# BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
# ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import datetime
import math
import weakref
import pygame
import numpy as np
import carla
from carla import ColorConverter as cc


def get_actor_display_name(actor, truncate=250):
    name = ' '.join(actor.type_id.replace('_', '.').title().split('.')[1:])
    return (name[:truncate-1] + u'\u2026') if len(name) > truncate else name


# ==============================================================================
# -- HUD -----------------------------------------------------------------------
# ==============================================================================

class HUD(object):
	def __init__(self, width, height):
		self.dim = (width, height)
		font = pygame.font.Font(pygame.font.get_default_font(), 20)
		fonts = [x for x in pygame.font.get_fonts() if 'mono' in x]
		default_font = 'ubuntumono'
		mono = default_font if default_font in fonts else fonts[0]
		mono = pygame.font.match_font(mono)
		self._font_mono = pygame.font.Font(mono, 14)
		self._notifications = FadingText(font, (width, 40), (0, height - 40))
		self.server_fps = 0
		self.frame = 0
		self.simulation_time = 0
		self._info_text = []
		self._server_clock = pygame.time.Clock()

	def on_world_tick(self, timestamp):
		self._server_clock.tick()
		self.server_fps = self._server_clock.get_fps()
		self.frame = timestamp.frame
		self.simulation_time = timestamp.elapsed_seconds

	def tick(self, world, ego, clock, showLabels=True):
		if ego.carlaActor is None:
			return  # ego not spawned yet

		t = ego.carlaActor.get_transform()
		v = ego.carlaActor.get_velocity()
		c = ego.carlaActor.get_control()

		heading = 'N' if abs(t.rotation.yaw) < 89.5 else ''
		heading += 'S' if abs(t.rotation.yaw) > 90.5 else ''
		heading += 'E' if 179.5 > t.rotation.yaw > 0.5 else ''
		heading += 'W' if -0.5 > t.rotation.yaw > -179.5 else ''

		#colhist = collisionSensor.get_collision_history()
		#collision = [colhist[x + self.frame - 200] for x in range(0, 200)]
		#max_col = max(1.0, max(collision))
		#collision = [x / max_col for x in collision]

		vehicles = world.get_actors().filter('vehicle.*')
		pedestrians = world.get_actors().filter('walker.pedestrian.*')

		self._info_text = [
			'Server:  % 16d FPS' % self.server_fps,
			'Map:	 % 20s' % world.get_map().name,
			'Simulation time: % 12s' % datetime.timedelta(seconds=int(self.simulation_time)),
			'',
			'Number of vehicles: % 8d' % len(vehicles),
			'Number of pedestrians: % 8d' % len(pedestrians),
			'',
			'Ego: % 20s' % get_actor_display_name(ego.carlaActor, truncate=20),
			'',
			'Speed:   % 15.0f m/s' % math.sqrt(v.x**2 + v.y**2 + v.z**2),
			u'Heading:% 16.0f\N{DEGREE SIGN} % 2s' % (t.rotation.yaw, heading),
			'Location:% 20s' % ('(% 5.3f, % 5.3f)' % (t.location.x, t.location.y)),
			'Height:  % 18.0f m' % t.location.z,
		]

		try:
			_control_text = [
				'',
				('Throttle:', c.throttle, 0.0, 1.0),
				('Steer:', c.steer, -1.0, 1.0),
				('Brake:', c.brake, 0.0, 1.0),
				('Reverse:', c.reverse),
				('Hand brake:', c.hand_brake),
				('Manual:', c.manual_gear_shift),
				'Gear:		%s' % {-1: 'R', 0: 'N'}.get(c.gear, c.gear),
				#'',
				#'Collision:', collision,
			]
		except:
			_control_text = []
		finally:
			self._info_text.extend(_control_text)

		if len(vehicles) > 1:
			self._info_text += ['Nearby vehicles:']
			distance = lambda l: math.sqrt((l.x - t.location.x)**2 + (l.y - t.location.y)**2 + (l.z - t.location.z)**2)
			vehicles = [(distance(x.get_location()), x)
						for x in vehicles if x.id != ego.carlaActor.id]
			for d, vehicle in sorted(vehicles):
				if d > 200.0:
					break
				vehicle_type = get_actor_display_name(vehicle, truncate=22)
				self._info_text.append('% 4dm %s' % (d, vehicle_type))

	def render(self, display):
		info_surface = pygame.Surface((220, self.dim[1]))
		info_surface.set_alpha(100)
		display.blit(info_surface, (0, 0))
		v_offset = 4
		bar_h_offset = 100
		bar_width = 106
		for item in self._info_text:
			if v_offset + 18 > self.dim[1]:
				break
			if isinstance(item, list):
				if len(item) > 1:
					points = [(x + 8, v_offset + 8 + (1.0 - y) * 30) for x, y in enumerate(item)]
					pygame.draw.lines(display, (255, 136, 0), False, points, 2)
				item = None
				v_offset += 18
			elif isinstance(item, tuple):
				if isinstance(item[1], bool):
					rect = pygame.Rect((bar_h_offset, v_offset + 8), (6, 6))
					pygame.draw.rect(display, (255, 255, 255), rect, 0 if item[1] else 1)
				else:
					rect_border = pygame.Rect((bar_h_offset, v_offset + 8), (bar_width, 6))
					pygame.draw.rect(display, (255, 255, 255), rect_border, 1)
					f = (item[1] - item[2]) / (item[3] - item[2])
					if item[2] < 0.0:
						rect = pygame.Rect((bar_h_offset + f * (bar_width - 6), v_offset + 8), (6, 6))
					else:
						rect = pygame.Rect((bar_h_offset, v_offset + 8), (f * bar_width, 6))
					pygame.draw.rect(display, (255, 255, 255), rect)
				item = item[0]
			if item: # At this point has to be a str
				surface = self._font_mono.render(item, True, (255, 255, 255))
				display.blit(surface, (8, v_offset))
			v_offset += 18


# ==============================================================================
# -- FadingText ----------------------------------------------------------------
# ==============================================================================

class FadingText(object):
	def __init__(self, font, dim, pos):
		self.font = font
		self.dim = dim
		self.pos = pos
		self.seconds_left = 0
		self.surface = pygame.Surface(self.dim)

	def set_text(self, text, color=(255, 255, 255), seconds=2.0):
		text_texture = self.font.render(text, True, color)
		self.surface = pygame.Surface(self.dim)
		self.seconds_left = seconds
		self.surface.fill((0, 0, 0, 0))
		self.surface.blit(text_texture, (10, 11))

	def tick(self, _, clock):
		delta_seconds = 1e-3 * clock.get_time()
		self.seconds_left = max(0.0, self.seconds_left - delta_seconds)
		self.surface.set_alpha(500.0 * self.seconds_left)

	def render(self, display):
		display.blit(self.surface, self.pos)


# ==============================================================================
# -- CollisionSensor -----------------------------------------------------------
# ==============================================================================

class CollisionSensor(object):
    def __init__(self, world, actor, hud=None):
        self.sensor = None
        self._history = []
        self._actor = actor
        self._actor_mass = actor.get_physics_control().mass
        self._hud = hud
        self._world = world
        bp = self._world.get_blueprint_library().find('sensor.other.collision')
        self.sensor = self._world.spawn_actor(bp, carla.Transform(), attach_to=self._actor)
        # Pass the lambda a weak reference to self to avoid circular reference
        weak_self = weakref.ref(self)
        self.sensor.listen(lambda event: CollisionSensor._on_collision(weak_self, event))

    def get_collision_speeds(self):
        '''Convert collision intensities from momentem (kg*m/s) to speed (m/s).'''
        return [(c[0], c[1] / self._actor_mass) for c in self._history]

    def get_collision_history(self):
        history = collections.defaultdict(int)
        for frame, intensity in self._history:
            history[frame] += intensity
        return history

    @staticmethod
    def _on_collision(weak_self, event):
        self = weak_self()
        if not self:
            return
        actor_type = get_actor_display_name(event.other_actor)
        if self._hud:
            self._hud.notification('Collision with %r, id = %d'
                                   % (actor_type, event.other_actor.id))
        impulse = event.normal_impulse
        intensity = math.sqrt(impulse.x ** 2 + impulse.y ** 2 + impulse.z ** 2)
        self._history.append((event.frame, intensity))
        if len(self._history) > 4000:
            self._history.pop(0)


# ==============================================================================
# -- CameraManager -------------------------------------------------------------
# ==============================================================================

class CameraManager(object):
	def __init__(self, world, actor, hud):
		self.sensor = None
		self._surface = None
		self._actor = actor
		self._hud = hud
		self.images = []
		self._camera_transforms = [
			carla.Transform(carla.Location(x=-5.5, z=2.8), carla.Rotation(pitch=-15)),
			carla.Transform(carla.Location(x=1.6, z=1.7))]
		self._transform_index = 1
		self._sensors = [
			['sensor.camera.rgb', cc.Raw, 'Camera RGB'],
			['sensor.camera.depth', cc.Raw, 'Camera Depth (Raw)'],
			['sensor.camera.depth', cc.Depth, 'Camera Depth (Gray Scale)'],
			['sensor.camera.depth', cc.LogarithmicDepth, 'Camera Depth (Logarithmic Gray Scale)'],
			['sensor.camera.semantic_segmentation', cc.Raw, 'Camera Semantic Segmentation (Raw)'],
			['sensor.camera.semantic_segmentation', cc.CityScapesPalette,
			 'Camera Semantic Segmentation (CityScapes Palette)'],
			['sensor.lidar.ray_cast', None, 'Lidar (Ray-Cast)']]
		self._world = world
		bp_library = self._world.get_blueprint_library()
		for item in self._sensors:
			bp = bp_library.find(item[0])
			if item[0].startswith('sensor.camera'):
				bp.set_attribute('image_size_x', str(hud.dim[0]))
				bp.set_attribute('image_size_y', str(hud.dim[1]))
			item.append(bp)
		self._index = None

	def toggle_camera(self):
		set_transform((self._transform_index + 1) % len(self._camera_transforms))

	def set_transform(self, idx):
		self._transform_index = idx
		self.sensor.set_transform(self._camera_transforms[self._transform_index])

	def set_sensor(self, index):
		index = index % len(self._sensors)
		needs_respawn = True if self._index is None \
			else self._sensors[index][0] != self._sensors[self._index][0]
		if needs_respawn:
			if self.sensor is not None:
				self.sensor.destroy()
				self._surface = None
			self.sensor = self._world.spawn_actor(
				self._sensors[index][-1],
				self._camera_transforms[self._transform_index],
				attach_to=self._actor)
			# Pass lambda a weak reference to self to avoid circular reference
			weak_self = weakref.ref(self)
			self.sensor.listen(lambda image: CameraManager._parse_image(weak_self, image))
		self._index = index

	def render(self, display):
		if self._surface is not None:
			display.blit(self._surface, (0, 0))

	@staticmethod
	def _parse_image(weak_self, image):
		self = weak_self()
		if not self:
			return
		if self._sensors[self._index][0].startswith('sensor.lidar'):
			points = np.frombuffer(image.raw_data, dtype=np.dtype('f4'))
			points = np.reshape(points, (int(points.shape[0] / 3), 3))
			lidar_data = np.array(points[:, :2])
			lidar_data *= min(self._hud.dim) / 100.0
			lidar_data += (0.5 * self._hud.dim[0], 0.5 * self._hud.dim[1])
			lidar_data = np.fabs(lidar_data)
			lidar_data = lidar_data.astype(np.int32)
			lidar_data = np.reshape(lidar_data, (-1, 2))
			lidar_img_size = (self._hud.dim[0], self._hud.dim[1], 3)
			lidar_img = np.zeros(lidar_img_size)
			lidar_img[tuple(lidar_data.T)] = (255, 255, 255)
			self._surface = pygame.surfarray.make_surface(lidar_img)
		else:
			image.convert(self._sensors[self._index][1])
			array = np.frombuffer(image.raw_data, dtype=np.dtype("uint8"))
			array = np.reshape(array, (image.height, image.width, 4))
			array = array[:, :, :3]
			array = array[:, :, ::-1]
			self._surface = pygame.surfarray.make_surface(array.swapaxes(0, 1))
		self.images.append(image)

	def destroy_sensor(self):
		if self.sensor is not None:
			self.sensor.stop()
			self.sensor.destroy()
